/*++++++++++++++
.IDENTIFICATION catutil.h
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Large Catalogues at CDS
.KEYWORDS       CDS Catalogue Server
.VERSION  1.0   02-Dec-2006
.COMMENTS       Generic definitions
---------------*/

#ifndef CATUTIL_DEF
#define CATUTIL_DEF	0

#ifndef _ARG
#ifdef __STDC__
#define _ARG(A)	A    /* ANSI */
#else
#define _ARG(A)	()  /* non-ANSI */
#define const
#endif
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
typedef int (*SORT_FCT)(const void *, const void *);

/* Verify 64-bit machine */
#ifndef int8
#define int8 long long
#endif
#ifndef int4
#define int4 int
#endif	/* int4 */
#ifndef int2
#define int2 short int
#endif
#ifndef int1
#define int1 unsigned char
#endif

#define ITEMS(a)	(sizeof(a)/sizeof(a[0]))
#define MIN(a,b)	((a)<(b) ? (a) : (b))
#define MAX(a,b)	((a)>(b) ? (a) : (b))
#define zero(a,l)	memset(a,0,l)
#define issign(c)	(((c)=='+')||((c)=='-'))

#define  sind(x)	sin((x)*(M_PI/180.))
#define  cosd(x)	cos((x)*(M_PI/180.))
#define  tand(x)	tan((x)*(M_PI/180.))
#define asind(x)	(180./M_PI)*asin(x)
#define atand(x)	(180./M_PI)*atan(x)
#define atan2d(x,y)  	(180./M_PI)*atan2(x,y)

#define adeg(x)		((x)*3600*1000)	/* Degree into mas */

#define nan2		0x8000		/* NaN for integer2 */
#define isnan2(x)	((x)&0xffff)==0x8000
#define nan4		0x80000000
#define isnan4(x)	(unsigned int)(x) == nan4

#if SWAP == 0
#define swap2(x)	(x)
#define swap4(x)	(x)
#define swap8(x)	(x)
#define swapa(a)	/* Nothing to swap an address */
#endif

#if SWAP == 1		/* SHORT swapping */
#define swap2(x)	((((x)&0xff)<<8)|(((x)&0xff00)>>8))
#define swap4(x)	((((x)&0xff00ff)<<8)|(((x)>>8)&0xff00ff))
#define swap8(x)	((((x)&0xff00ff00ff00ffLL)<<8)\
			|(((x)>>8)&0xff00ff00ff00ffLL))
#define swapa(a)	swap_a1((unsigned char *)a, sizeof(*a))
#endif

#if SWAP == 2		/* FULL swapping */
#define swap2(x)	((((x)&0xff)<<8)|(((x)&0xff00)>>8))
#define swap4(x)	((((x)&0xff)<<24)|(((x)&0xff00)<<8)\
			|(((x)>>24)&0xff)|(((x)>>8)&0xff00))
#define swap8(x)	((((x)&0xff)<<56)|(((x)&0x00ff00)<<40)\
			|(((x)>>56)&0xff)|(((x)>>40)&0x00ff00)\
			|(((x)&0x00ff0000)<<24)|(((x)&0xff000000)<<8)\
			|(((x)>>24)&0x00ff0000)|(((x)>>8)&0xff000000))
#define swapa(a)	swap_a2((unsigned char *)a, sizeof(*a))
#endif

#if SWAP == 3		/* Impossible swapping */
#define swap2(x)	ERROR SWAPPING
#define swap4(x)	ERROR SWAPPING
#endif

/*===========================================================================
		Prototypes Declarations
 *===========================================================================*/

void  verify_swap _ARG((void));
void  swap_a1     _ARG((unsigned char *a, int size));

int   strloc   _ARG((char *text, int c));	/* similar to   strchr */
char *strdup2  _ARG((char *t1, char *t2));	/* new string t1 + t2  */
int   get_bits _ARG((unsigned char *a, int offset_bits, int len_bits));
int   isqrt    _ARG((int n));			/* Integer square root */
int   ihypot   _ARG((int a, int b));		/* Integer distance    */

/* Locate index | start[i] <= value < start[i+1] */
int   binloc     _ARG((int4 *start, int items, int4 value));
/* Locate index | f(p) <= value < f(p+reclen)    */
int   binlocf  _ARG((char *start, int bytes, int reclen, int4 (*f)(),
	    int4 value));
int   abinloc  _ARG((void **start, int items, void **new, SORT_FCT cmp));

void  tr_ou  _ARG((double o[2], double u[3]));
void  tr_uo  _ARG((double u[3], double o[2]));
void  tr_oR  _ARG((double o[2], double R[3][3]));

/* Rotate        v = R u */
void  tr_uu  _ARG((double u[3], double v[3], double R[3][3]));
/* Rotate        v = tR u */
void  tr_uu1 _ARG((double u[3], double v[3], double R[3][3]));

/* Interpretation of strings */
int   match_int4   _ARG((char *string, int4 *value));
/* Interpret an unsigned number, possibly e-notation */
int   match_unum   _ARG((char *string, double *value, int *ndec));
/* Interpret a  number, possibly e-notation */
int   match_num    _ARG((char *string, double *value));
/* Interpret 2 numbers */
int   match_num2   _ARG((char *string, double values[2]));
/* Interpret a string for a Jname */
int   match_Jname  _ARG((char *arg, int4 *mas, int4 *Jprec));
/* Interpret a center */
int   match_center _ARG((char *string, double center[2]));

/* Editions in Sexagesimal */
int   edsex_ra  _ARG((char *buffer, int4 ra_mas, int ndec));
int   edsex_de  _ARG((char *buffer, int4 de_mas, int ndec));
int   eddeg_ra  _ARG((char *buffer, int4 ra_mas, int ndec));
int   eddeg_de  _ARG((char *buffer, int4 de_mas, int ndec));
int   edmas_ra  _ARG((char *buffer, int4 ra_mas));
int   edmas_de  _ARG((char *buffer, int4 de_mas));

#endif	/* CATUTIL_DEF */
