/*++++++++++++++
.IDENTIFICATION ppmxl.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    PPMXL Catalogue
.KEYWORDS       CDS Catalogue Server
.VERSION  0.9   10-May-2010:
.VERSION  1.0   28-Jun-2010
	Remember that the environment variable PPMXLroot provides
	the root of PPMXL compressed catalogue; its default is
	$PPMXLroot (/cats/PPMXL)
---------------*/

#define VERSION "1.0"

#ifndef SORT_METHOD
#define SORT_METHOD 1	/* Use array of pointer */
#endif

#include <ppmxl.h>	/* Includes stdlib, stdio, string, ctype, math */
#include <string.h>
#include <unistd.h>

#ifndef ROOTdir		/* Normally value supplied in compilation */
#define ROOTdir "/cats/PPMXL"
#endif

#ifndef BINSIZE		/* Size of blocks in binary output */
#define BINSIZE		500
#endif

#ifndef TIMEOUT
#define TIMEOUT 600
#define MAX_TIMEOUT (99999)
#endif
static int4 timeout = 0;
static time_t started;

#define RADIUS		(5/60.)		/* Default target radius */

#if 0	/* defined in catutil.h */
#define adeg(x)		((x)*3600*1000)	/* Degree into mas */
#endif
#define Puts(s)		fprintf(ofile, "%s\n", s)

/* Options for PPMXL queries.......	*/
int ppmxl_options = 0 ;	/* 1=verbose, 0x10 for Original Order
			 * 0x100 = Binary output
                        */
static unsigned int photomask = 0 ; /* Requires photometry */
static char *theProg;

#if SORT_METHOD == 2
/* Linked records, to sort them...	*/
typedef struct linked_PPMXL {
    struct linked_PPMXL *lt ;
    struct linked_PPMXL *eq ;
    struct linked_PPMXL *gt ;
    PPMXLrec rec ;
} LPPMXLrec ;
static LPPMXLrec *last, *arec, *root;
#else
/* Use an array of pointers */
#define LPPMXLrec PPMXLrec
static PPMXLrec *arec;
static PPMXLrec **root;
#endif
#define null_mag(v)  ((v) == NULL2)

static FILE *ofile;
static int binlen = sizeof(PPMXLrec)-20; /* r theta x y pmt not transmitted */
static unsigned char *oblock;	/* Blocked binary output */
static int     iblock = 0;	/* Number of records     */
static int     mblock = 0;	/* Max-# of records      */
static int     kblock = 0;	/* Indicator ^B written  */

static int4 compare_calls;
static int mrec = BINSIZE ;
static int irec, truncated ;
static char input_id ;		/* Type of Input:  */
static char *prompt[] = {
    "Center",		/* input_id=0 */
    "PPMXL-ID",		/* input_id=1 */
    "RA+Dec Limits",	/* input_id=2 */
    "Jname",		/* input_id=3 */
    (char *)0
};
static char whole_ppmxl ; /* -whole option */

/*  Definitions related to Center of Field */
static double radius = 0 ;
static double boxy[2] ;
static double localR[3][3] ;	/* Matrix, [0] = dir.cos. of center */
static double du2max = -1 ;

/* Other options */
static int  opted = 6 ; 	/* Default edit Deg, Epoch, ID */
static int  optE = 0;

/* Definitions of the record fields */
struct s_fields {
    char name[9];
    char selected ;
    char order ;	/* 1 = descending */
    char null ;		/* 2 = may be 0x8000 for NULL */
    unsigned int4 photomask;	/* Option bits for ppmxl_search */
    double factor ;	/* 0: ascii */
    int4 lim[2] ;
};
static struct s_fields thefields[99] =  {
    {"r",0,0,0,0,3.6e6},		/* MUST BE #0 */
    {"i",0,0,0,0,1.},			/* MUST BE #1 */
    {"x",0,0,0,0,(180./M_PI)*3.6e6},	/* MUST BE #2 */
    {"y",0,0,0,0,(180./M_PI)*3.6e6},	/* MUST BE #3 */
    {"a",0,0,0,0,3.6e6},		/* Must be #4 */
    {"d",0,0,0,0,3.6e6},		/* Must be #5 */
#define magFIELD 6
    {"mJ",0,0,2,0x80,1000.},		/* Must be #6 */
    {"mH",0,0,2,0x40,1000.},		/* Hmag  */
    {"mK",0,0,2,0x20,1000.},		/* Ksmag */
    {"mb1",0,0,2,0x10,1000.},
    {"mb2",0,0,2,0x08,1000.},
    {"mr1",0,0,2,0x04,1000.},
    {"mr2",0,0,2,0x02,1000.},
    {"mi",0,0,2,0x01,1000.},
    {"mb",0,0,2,0x118,1000.},		/* 0x100: b1 OR b2 */
    {"mr",0,0,2,0x206,1000.},		/* 0x200: r1 OR r2 */
    {"p",0,0,0,0,10.},		/* Total proper motion */
    {"ep",0,0,0,0,100.},		/* Epoch */
    {"ea",0,0,0,0,1.},		/* error(RA) , mas */
    {"ed",0,0,0,0,1.},		/* error(Dec), mas */
    {"n",0,0,0,0,1.},		/* Nobs  */
    {"s",0,0,0,0,1.},		/* Survey flags 0..15 */
    {"", 0,0,2,0,1000.},	/* Here free fields for colors... */
    /* {"\377\377\377\377"} 	-- last field	*/
} ;
static struct s_fields *compare_fields[32] ;
static struct s_fields *check_fields[32] ;
static int4 matched ;	/* Matching records 	*/
static int head_issued = 0;
static char ppmxl_title[] = "\
#======== PPMXL server (2010-06, V%s) ======== CDS, Strasbourg ========\n\
";
static char optEmsg[] = "\
#---------Position recomputed at Epoch.";

static char usage[] = "\
Usage: ppmxl [-HELP] [-R root_name] [-E] [-r[s] [min,]radius] [-b[s] x[,y]]\n\
             [-e edit_opt] [-f input_file] [-m max_records] [-obin file|-]\n\
	     [-c center | -i ID | Jname | -whole] [-l! min,max] [-s !]\n\
";

static char help[] = "\
  -HELP: display column explanations, -HELP-bin the binary structure\n\
     -E: Compute the position at mean Epoch (apply the proper motion back)\n\
     -b: target box in arcmin ;       -bs = target box in arcsec\n\
     -r: target radius in arcmin (5); -rs = target radius in arcsec (300)\n\
     -c: target center in decimal or sexagesimal (defaut in stdin)\n\
     -e: s=edit Pos in Sexa, m=mas; i=edit ID, e=edit ObsDate, x=edit x,y\n\
     -f: specifies an input file (default stdin)\n\
     -i: choice of ID (18-digit) instead of a position\n\
     -m: max number of stars to retrieve\n\
  -obin: write output in binary mode (byte order as FITS) to specified file\n\
    -l!: Set the limits (range) on one of the parameters (below)\n\
    -s!: Sort the result by the parameter ! (list below)\n\
 -whole: search on the whole sky\n\
====The abbreviations of the parameters (symbolized !)are:\n\
      a=Alpha  d=Delta       e=Epoch     i=PPMXL-name  m=anyMagOrColor\n\
      n=Nobs   p=pm(mas/yr)  s=status    r=distance   x=proj.E y=proj.N\n\
'm' may be followed by the filter J H K (or lowercase) b1 b2 r1 r2 i\n\
        and eventually by a second magnitude as e.g. mi-J\n\
";

static char help_bin[] = "\
#In binary mode (option -obin), the output stream starts by a Control-B char\n\
#followed by buffers preceded by a 4-byte integer representing the length;\n\
#the end of the binary stream is indicated by a zero-length buffer.\n\
#The first buffer contains a single record (it gives the length of a record);\n\
#the following ones may be blocked with an arbitrary factor.\n\
#All binary data are give in network byte order (most significant byte first)\n\
";

static char HELP[] = "\
PPMXL Catalog (PPM-Extended, positions and proper motions by Roeser et al. 2008)\n\
=============================================================================\n\
\n\
The fields have the following meaning:\n\
\n\
PPMXL      = Designation from ICRS position (HHMMSS.s+DDMMSS);\n\
            it may be followed by a 'p' (preceding) of 'f' (following)\n\
RA+Dec    = Position (Alpha, Delta) in ICRS (Hipparcos system, more accurate\n\
            but compatible with J2000)\n\
pmRA,pmDE = proper motions along RA and Dec directions (mas/yr)\
epRA,epDE = Epochs to which the errors on the positions are valid\n\
eRA,eDE   = mean errors on Right Ascension and Declination (mas, at epoch)\n\
epma,epmd = mean errors on proper motion components (mas/yr)\n\
Cmag      = Catalog magnitude (from source catalog)\n\
Rmag      = Derived red magnitude (579-642nm)\n\
Bmag +/-  = Johnson blue magnitude, and mean error in milli-mag\n\
Vmag +/-  = Johnson visual magnitude, and mean error in milli-mag\n\
Jmag +/-  = Near-IR J magnitude (1.2um) from 2MASS catalog\n\
Hmag +/-  = Near-IR J magnitude (1.6um) from 2MASS catalog\n\
Kmag +/-  = Near-IR K magnitude (2.6um) from 2MASS catalog\n\
No,P      = Number of observation in LSQ fit, 'P' indicates a problematic fit\n\
S         = Status flags: 1=large chi, 2/4=PPMX row, 8=merged USNO roww\n\
Source ID = Identificaiton in source catalog where:\n\
            H=ARIHIP, T=Tycho-2, P=PPM, G=GSC1.2, A=AC2000\n\
r(\")     = distance from specified center, in arc seconds.\n\
=============================================================================\n\
";

/* argument in range [-180,180[ */
#define mrange(x)  if((x)>=adeg(180)) x -= adeg(180); \
	      else if((x)<-adeg(180)) x += adeg(180);
/*==================================================================
		Get the parameters
 *==================================================================*/


static struct s_fields *dup_field(struct s_fields *field)
/*++++++++++++++++
.PURPOSE  Clone a field
.RETURNS  The copy
-----------------*/
{
  struct s_fields *f;
    f = (struct s_fields *)malloc(sizeof(struct s_fields));
    memcpy(f, field, sizeof(struct s_fields)) ;
    return(f);
}

static int mag_value(short int *amag, char *name)
/*++++++++++++++++
.PURPOSE  Interpret the filter name as index
.RETURNS  The index / -1 if error
-----------------*/
{
  int i, val=0, sub=0;	/* sub=1 if subtraction asked */
  locate_mag:
    if (name[0] == 'm') name++;
    i = strloc(ppmxl_col, *name);
    if (ppmxl_col[i] == 0) return(NULL2);
    name++;  /* gives '1' '2' '-' or 0 */
    if (isdigit(*name)) {
	if (*name == '2') i++;
	name++;
    }
    else if ((i>=3)&&(i<=6)) {
	if (null_mag(amag[i])) i++;
    }
    if (null_mag(amag[i])) return(NULL2);
    if (sub) val -= amag[i]; else val += amag[i];

    /* Find 2nd magnitude */
    if (*name != '-') 	/* Not a color index */
        return(val);
    name++; sub=1;
    goto locate_mag;
}

static struct s_fields *get_field(struct s_fields *fields, char *name)
/*++++++++++++++++
.PURPOSE  Interpret the meaning of a parameter
.RETURNS  The relevant field / NULL
-----------------*/
{
  struct s_fields *f, *f1, *f2;
  int letter, i, j;
  char *p;
    letter = tolower(name[0]);
    f1 = (struct s_fields *)0;
    for (f=fields; f->name[0]; f++) {
	if (f->name[0] != letter) continue;
	for (j=1; name[j] && (tolower(f->name[j])==tolower(name[j])); j++) ;
	if (f->name[j] == 0) {
	    if (name[j] == 0) return(f);	/* Exact match */
	    if (name[j] == '-') f1 = f;		/* 1st color match */
	}
    }
    if (f[1].name[0] == '\377') { 	   /* Very last free field */
	fprintf(stderr, "#***Too many fields, ignored: %s\n", name);
	 return((struct s_fields *)0);
    }
    if (!f1)
	goto nosuch_parameter;

    /* Incomplete match. Here:
     * f  indicates the first free condition field
     * f1 indicates the first matched color (e.g. mJ)
     */

    /* try to match the second color */
    i = strlen(f1->name);
    if (name[i] == '-') i++;    /* In principle always true */
    if (name[i] == 'm') i++;    /* Accept e.g. mj-mb        */
    p = name + i-1;   		/* p[1] is letter of band   */
    for (f2=fields; f2->name[0]; f2++) {
	if (f2->name[0] != 'm') continue;
	for (j=1; p[j] && (tolower(f2->name[j])==tolower(p[j])); j++) ;
	if (p[j] || f2->name[j]) continue;
	/* Color matched, create the new field type */
	memcpy(f, f1, sizeof(struct s_fields));
	strcat(f->name, "-");
	strcat(f->name, f2->name+1);
	f->photomask |= f2->photomask;
	return(f);
    }
  nosuch_parameter:
    fprintf(stderr, "#***Can't understand field '%s'\n", name);
    return((struct s_fields *)0);
}

#if 0
static int get_lim (char *string, double values[2])
/*++++++++++++++++
.PURPOSE  Find out the limits (2 numbers separated by ,)
.RETURNS  0 / 1 / 2 / 3 / -1=error
.REMARKS  Accept p=min1,max1=min2,max2 ==> sqrt(min1^2+min2^2)
-----------------*/
{
  int found=0;
  int n;

    n = match_num2(string, values);
    if (string[n]) {	/* Not totally matched ? */
	fprintf(stderr, "****'%s' not understood in limits: %s\n",
		string+n, string);
	return(-1);
    }
    found =  values[0] == values[0];	/* a true number found*/
    if (values[1] == values[1]) found |= 2;

    return(found);
}
#endif

/*==================================================================
		Check routine (return 0 = don;t keep, 1 = keep
 *==================================================================*/

static int remove_pm(PPMXLrec *rec)
/*++++++++++++++++
.PURPOSE  Remove the proper motion (compute position at Epoch)
.RETURNS  0 (no proper motion) / 1
.REMARKS  Use matrix for position close to pole
   dx    -sin(a) -sin(d)cos(a)
   dy  =  cos(a) -sin(d)sin(a)  .  mu1.t
   dz       0    +cos(d)           mu2.t
-----------------*/
{
  double R[3][3], u[3], v[3], o[2], t;
  int ra, de;
    if (rec->fill&1)  /* Already done */
    if ((rec->pmra | rec->pmde) == 0)   /* Zero proper motion   */
        return(0);
    rec->fill |= 1;
    t = (rec->epra + rec->epde - 400000)/200.;
    if (t == 0) return(0);              /* E.g. for Tycho-2     */
    o[0] = rec->ra; o[1] = rec->de;
    o[0] /= 3.6e6; o[1] /= 3.6e6;
    v[1] = rec->pmra*t/3.6e7;
    v[2] = rec->pmde*t/3.6e7;
    if (fabs(o[1]) > 84) {              /* Close to Pole */
        v[0] = 1;
        v[1] *= (M_PI/180.);
        v[2] *= (M_PI/180.);
        tr_oR(o, R);
        tr_uu1(v, u, R);
        tr_uo(u, o);
    }
    else {
        o[0] += v[1]/cosd(o[1]);
        if (o[0] <    0.) o[0] += 360.;
        if (o[0] >= 360.) o[0] -= 360.;
        o[1] += v[2];
    }
    /* Convert to mas */
    ra = adeg(o[0]); de = adeg(o[1]);
#if 0   /* How large is the change ?? */
    fprintf(stderr, "....Change for %4d-%07d (%+05.1fyr) %6d %6d\n", 
      rec->zone, rec->id, t, ra-rec->ra, de-rec->de);
#endif
    rec->ra = ra; rec->de = de;
    return(1);
}

static int check(PPMXLrec *rec)
/*++++++++++++++++
.PURPOSE  Verify the various constraints
.RETURNS  1 (OK) / 0 (does not fit)
-----------------*/
{
  struct s_fields **af, *f;
  double o[2], u[3], du, dr2 ;
  int i=0, val, flag=0;  /* flag set to 1 if r computed, 2 if (xy) computed */

    /* Verify first the Position */
    rec->xy[0] = rec->xy[1] = NULLxy ;
    if (du2max >= 0) {
    	o[0] = rec->ra / 3.6e6 ;
    	o[1] = rec->de / 3.6e6 ;
    	tr_ou(o, u); flag |= 1 ;
    	du = u[0] - localR[0][0] ; dr2  = du * du ;
    	du = u[1] - localR[0][1] ; dr2 += du * du ;
    	du = u[2] - localR[0][2] ; dr2 += du * du ;
    	if (dr2 > du2max)
	    return(0) ;
    	rec->rho = thefields[0].factor * 2.*asind(0.5*sqrt(dr2));
	if (opted&8) {	/* Compute x,y */
      	    tr_uu(u, u, localR) ;
      	    rec->xy[0] = thefields[2].factor*u[1]/u[0] ;
      	    rec->xy[1] = thefields[3].factor*u[2]/u[0] ;
      	    flag |= 2 ;
	}
    }
    else rec->rho = -1 ;

    /* printf("#id=%4d,%8d, rho=%8d, mags=%4d,%4d,%4d\n",
	rec->zone,rec->id,rec->rho, rec->mB, rec->mR, rec->ci) ;
    */

    for (af=check_fields ; *af; af++) {
	f = *af;
	switch(f->name[0]) {
	  case 'r':		/* Distance to center */
	    if (rec->rho < f->lim[0]) return(0);
	    if (rec->rho > f->lim[1]) return(0) ;
	    continue ;
          case_xy:
      	    if (!flag) {
    	        o[0] = rec->ra / 3.6e6;
    	        o[1] = rec->de / 3.6e6;
    	        tr_ou(o, u); flag |= 1 ;
      	    }
      	    if (!(flag&2)) {
      	        tr_uu(u, u, localR) ;
      	        rec->xy[0] = thefields[2].factor*u[1]/u[0] ;
      	        rec->xy[1] = thefields[3].factor*u[2]/u[0] ;
      	        flag |= 2 ;
      	    }
	    if (rec->xy[i] < f->lim[0]) return(0) ;
	    if (rec->xy[i] > f->lim[1]) return(0) ;
      	    break ;
          case 'x': i = 0 ; goto case_xy ;
          case 'y': i = 1 ; goto case_xy ;
          case 'a':
	    if (f->lim[0] <= f->lim[1]) {
	        if (rec->ra < f->lim[0]) return(0) ;
	        if (rec->ra > f->lim[1]) return(0) ;
	    }
	    else {
	        if ((rec->ra < f->lim[0]) &&
	            (rec->ra > f->lim[1])) return(0) ;
	    }
	    continue ;
	  case 'd':			/* Declinaison */
	    if (rec->de < f->lim[0]) return(0) ;
	    if (rec->de > f->lim[1]) return(0) ;
	    continue ;
	  case 'p':			/* Total proper motion */
	    if (rec->pmt < 0) rec->pmt = ihypot(rec->pmra, rec->pmde);
	    val = rec->pmt;
	    goto case_val;
	  case 'm':			/* Magnitude / color */
	    val = mag_value(rec->mag, f->name);
	    /* NO BREAK */
	  case_val:
	    if ((f->null) && null_mag(val))	/* NaN */
		return(0);
	    if (val < f->lim[0]) return(0);
	    if (val > f->lim[1]) return(0);
	    continue;
	  case 'e':			/* Epoch */
	    val = (rec->epra + rec->epde)/2;
	    goto case_val;
          case 'n':			/* Number of observations */
	    i = rec->Nobs;
	    if (i < f->lim[0]) return(0) ;
	    if (i > f->lim[1]) return(0) ;
	    continue ;
	  case 's':			/* Status bits */
	    /* lim[0] = list of bits that MUST be zero 
	     * lim[1] = list of bits that MUST be one */
	    if (rec->flags&f->lim[0]) return(0);
	    if((rec->flags&f->lim[1]) != f->lim[1]) return(0);
	    continue;
	  case 'i':			/* Should never happen... */
	    fprintf(stderr, "#***ppmxl_check: no test on '%s'\n", f->name);
	    return(-1);
	}
    }
    return(1) ;
}

/*==================================================================
		Comparison routines
 *==================================================================*/

static int compare(PPMXLrec *a, PPMXLrec *b)
{
  struct s_fields *f, **af;
  int diff=0, val;
    compare_calls++;
    for (af=compare_fields ; (diff==0) && *af; af++) {
        f = *af;
        switch(f->name[0]) {
          case 'r':
            diff = a->rho - b->rho;             break;
          case 'x':
            diff = a->xy[0] - b->xy[0];         break;
          case 'y':
            diff = a->xy[1] - b->xy[1];         break;
          case 'a':
            diff = a->ra - b->ra;               break;
          case 'd':
            diff = (a->de - b->de);             break;
          case 'p':
	    if (a->pmt<0) a->pmt = ihypot(a->pmra, a->pmde);
	    if (b->pmt<0) b->pmt = ihypot(b->pmra, b->pmde);
            diff = (a->pmt - b->pmt);           break;
          case 'm':             /* Magnitudes + Colors */
	    val = mag_value(a->mag, f->name);
            if (null_mag(val)) return(1);
	    diff = val;
	    val = mag_value(b->mag, f->name);
            if (null_mag(val)) return(-1);
            diff -= val;
            break;
          case 'e':                     /* Epoch */
            diff = ((a->epra - b->epra) + (a->epde - b->epde))/2;
            break;
	  case 's':
	    diff = (a->flags - b->flags);	break;
        }
        if (f->order) diff = -diff;
    }
    return(diff);
}

/*==================================================================
		Sorting the Records
 *==================================================================*/

#if SORT_METHOD == 2		/* Use Btree */
static int add_rec(PPMXLrec *new)
/*++++++++++++++++
.PURPOSE  Add a new record, link it.
.RETURNS  0 (not kept) / 1
-----------------*/
{
  LPPMXLrec *node, *prev, *n;
  int4 comp1;
  int diff;

    comp1 = compare_calls;

    /* (0) Check if the new record has any use... */
    n = (LPPMXLrec *)0;             /* n -> last */
    if (irec == mrec) {
        if (!last) for (last=root; last->gt; last=last->gt) n = last;
        diff = compare(new, &(last->rec));
        if (diff >= 0) { truncated++; return(0); }
    }

    /* (1) Find where in the list we've to insert the new record */
    node = root;
    prev = (LPPMXLrec *)0;
    while(node) {
        diff = compare(new, &(node->rec));
        prev = node;
        if (diff == 0) break;
        node = diff < 0 ? node->lt : node->gt ;
    }

    /* (2) If max attained, put the new in place of last */
    if (irec == mrec) {
        truncated++ ;
        node = last ;           /* Where to store new record    */
        if (last->eq) {         /* Several last records...      */
            node = last->eq;
            last->eq = node->eq ;
        }
        else {                  /* n->gt must give LAST record  */
            if (!n) for (last=root; last->gt; last=last->gt) n = last;
            if (!n) {           /* Happens when root == last !  */
                root = root->lt;
                last = root;
            }
            else {              /* The last is now given by n   */
                last = n;
                last->gt = node->lt;
            }
        }
        while(last->gt) last = last->gt;
    }
    else node = arec + irec++ ;

    /* (3) Insert the new record, and set the links */
    memcpy(&(node->rec), new, sizeof(PPMXLrec));
    /* *(&(node->rec)) = *new ; */
    node->lt = node->gt = node->eq = (LPPMXLrec *)0;
    if (!prev) { root = node; last = (LPPMXLrec *)0; }
    else {
        if (diff < 0) prev->lt = node;
        else if (diff > 0) {    /* I may have prev == last ...  */
            prev->gt = node;
            if (last) while(last->gt) last = last->gt;
        }
        else { node->eq = prev->eq; prev->eq = node; }
    }

#if 0
    fprintf(stderr, "....Adding#%d: comparisons=%ld\n",
        irec, compare_calls-comp1);
#endif

    return(1) ;
}
#else		/* Use array of pointers */
static int acompare(PPMXLrec **a, PPMXLrec **b)
/*++++++++++++++++
.PURPOSE  Comparison function to sort the array of PPMXL records
.RETURNS  Difference *a - *b
-----------------*/
{
    return(compare(*a, *b));
}
static void asort(PPMXLrec **root, int items)
/*++++++++++++++++
.PURPOSE  Sort the 'root' array
.RETURNS  0 (not kept) / 1
-----------------*/
{
  PPMXLrec **a, *p; int i;
    for (a=root, p=arec, i=items; --i>=0; p++, a++)
	*a = p;
    qsort(root, items, sizeof(PPMXLrec *), (SORT_FCT)acompare);
}

static int add_rec(PPMXLrec *new)
/*++++++++++++++++
.PURPOSE  Add a new record, sort if necessary
.RETURNS  0 (not kept) / 1
-----------------*/
{
  PPMXLrec *rec, **a, **alow;
  int i;

    /* (0) If there is room, just add it */
    if (irec < mrec) {
	memcpy(arec+irec, new, sizeof(PPMXLrec));
	irec++;
	return(1);
    }

    truncated++;	/* There are more records than acceptable */

    /* Sort if not done */
    if (!*root)
	asort(root, mrec);

    /* Check if the new record has any use... */
    a = root + mrec-1;
    if (compare(new, *a) >= 0)
	return(0);

    /* Store the new record in place of the last record */
    rec = *a;
    memcpy(rec, new, sizeof(PPMXLrec));

    /* Find where in the list we've to insert the new record */
    i = abinloc((void **)root, mrec-1, (void **)a, (SORT_FCT)acompare);
    alow = root+i;	/* i in range -1 .. mrec-1 */
    for (a = root+mrec-2; a>alow; a--) a[1] = a[0];
    alow[1] = rec;

    return(1) ;
}
#endif

/*==================================================================
		Writing the binary records
 *==================================================================*/
int ppmxl_write(PPMXLrec *pr)
/*++++++++++++++++
.PURPOSE  Write the (binary) PPMXL record (length + record)
.RETURNS  Number of bytes written
-----------------*/
{
  static char ctrl_b = '\02';
  unsigned char *abuf = 0;
  int4 length, pos[2];
  int  use_block = 1;
  int  bytes = 0;

    /* Write the Control-B prefix and the length */
    if ((kblock&1) == 0) {
	/* Compute the actual max# of records in a block */
	if (kblock == 0) {
	    if ((opted&(~17)) == 256)   /* Astrometry only */
		binlen = 8;
	    mblock /= binlen; 		/* Actual #records */
	    fprintf(ofile,
		    "#(^B starts binary output) binlen=%d max.rec=%d\n",
		    binlen, mblock);
	    mblock *= binlen; 		/* Actual #bytes   */
	}
	bytes += fwrite(&ctrl_b, 1, 1, ofile);
	kblock++;
	use_block = 0;
    }

    /* BLock the output for better efficiency */

    if (binlen == 8) {
	pos[0] = swap4(pr->ra);
	pos[1] = swap4(pr->de);
	abuf = (unsigned char *)pos;
    }
    else {
	ppmxl_swap(pr);
	abuf = (unsigned char *)pr;
    }
    if (use_block) {
	if (iblock == mblock) {
	    length = swap4(iblock);
	    bytes += fwrite(&length, 1, 4, ofile);
	    bytes += fwrite(oblock, 1, iblock, ofile);
	    iblock = 0;
	}
	memcpy(oblock+iblock, abuf, binlen);
	iblock += binlen;
    }
    else {	/* Write a single record */
	length = swap4(binlen);
	bytes += fwrite(&length, 1, 4, ofile);
        bytes += fwrite(abuf, 1, binlen, ofile);
    }
    if (abuf == (unsigned char *)pr)
	ppmxl_swap(pr);		/* Swap back    */
    return(bytes);
}

int ppmxl_flush()
/*++++++++++++++++
.PURPOSE  Flush (terminate the binary output)
.RETURNS  Number of bytes written
-----------------*/
{
  int bytes = 0;
  int4 length;

    if (kblock&1) {
        if (iblock>0) {
	    length = swap4(iblock);
	    bytes += fwrite(&length, 1, 4, ofile);
	    fwrite(oblock, 1, iblock, ofile);
	    iblock = 0;
	}
	length = 0;
	bytes += fwrite(&length, 1, 4, ofile);
	kblock++;
    }
    fflush(ofile);
    return(bytes);
}

#if SORT_METHOD == 2
static void print_result(LPPMXLrec *node)
/*++++++++++++++++
.PURPOSE  Print all nodes in order
.RETURNS  ---
.REMARKS  Recursivity = simplicity !!
-----------------*/
{
  LPPMXLrec *n;
    if (!node) return;
    if (node->lt) print_result(node->lt);
    for (n=node; n; n=n->eq) {
	if (ppmxl_options&0x100) 	/* Binary output */
	     ppmxl_write(&(n->rec));
	else Puts(ppmxl_a(&(n->rec), opted));
    }
    fflush(ofile);
    if (node->gt) print_result(node->gt);
}
#else
static void print_result(PPMXLrec **root)
/*++++++++++++++++
.PURPOSE  Print all records in order
.RETURNS  ---
.REMARKS  Write by blocks in binary output
-----------------*/
{
  PPMXLrec **a;
  int i;

    /* Sort if not done */
    if (!*root)
	asort(root, irec);

    /* Edit the values  */
    for (a=root, i=irec; --i>=0; a++) {
	if (ppmxl_options&0x100) 	/* Binary output */
	     ppmxl_write(*a);
	else Puts(ppmxl_a(*a, opted));
    }
}
#endif

static int digest(PPMXLrec *rec)
/*++++++++++++++++
.PURPOSE  Digest routine: check and display if necessary
.RETURNS  0
-----------------*/
{
  int st ;
    /* --- Apply the Position Transformation if needed */
    if (optE) remove_pm(rec);

    st = check(rec) ;
    if (st <= 0) return(st) ;
    matched += 1 ;
    if (compare_fields[0])
	add_rec(rec)  ;
    else if (ppmxl_options&0x100) {	/* Binary output */
	ppmxl_write(rec);
    }
    else {
	if (head_issued == 0) fprintf(ofile, "%s\n", ppmxl_head(opted));
	head_issued++;
	if (irec >= mrec) { truncated++; matched = 0; return(-1) ; }
	irec++ ;
    	fprintf(ofile, "%s\n", ppmxl_a(rec, opted));
    }
    return(0) ;
}

int4 ppmxl_loop()
/*++++++++++++++++
.PURPOSE  Loop on read & test of PPMXL records
.RETURNS  Number of tested records.
-----------------*/
{
  int4 tested = 0 ;
  PPMXLrec rec ;
    while(1) {
        if (ppmxl_read(&rec) <= 0) break ;
        tested++;
        if (digest(&rec) < 0) break ;
    }
    return(tested) ;
}

#if 1
int4 ppmxl_pos(int4 *gra, int4 *gde)
/*++++++++++++++++
.PURPOSE  Launch Search on PPMXL stars from limits in RA(gra[2]) + DE(gde[2])
.RETURNS  Number of tested records.
.REMARKS  Merge with specified RA / DE limits
-----------------*/
{
  int4 ra[4], de[2], rac[2], dra[2], qra[2], *ara, t, tested;
  int i;

    if (gde) { de[0] = gde[0],    de[1] = gde[1] ; }
    else     { de[0] = adeg(-90), de[1] = adeg(90); }
    if (thefields[5].selected) {	/* Limits in DE */
	de[0] = MAX(de[0], thefields[5].lim[0]) ;
	de[1] = MIN(de[1], thefields[5].lim[1]) ;
	if (de[0] > de[1]) return(0);	/* Mismatch Dec */
	gde = de;
    }

    if (!thefields[4].selected)
	return(ppmxl_search(gra, gde, photomask, digest));

    /* Incoporate the range in RA */
    ra[0] = thefields[4].lim[0];
    ra[1] = thefields[4].lim[1];
    if (!gra) 
	return(ppmxl_search( ra, gde, photomask, digest));
    if ((ra[0] == gra[0]) && (ra[1] == gra[1]))
	return(ppmxl_search( ra, gde, photomask, digest));

    /* Here 2 ranges in RA: ra[0,1] from thefields, ra[2,3] from argument */
    ra[2] = gra[0]; ra[3] = gra[1];
    for (i=0; i<4; i++) { if(ra[i]>0) ra[i] += adeg(360); }
    
    /* Compute center and semi-amplitude */
    for (i=0, ara=ra; i<2; i++, ara+=2) {
        rac[i] = ara[0]/2 + ara[1]/2 + ((ara[0]&1)+(ara[1]&1))/2;
        dra[i] = (ara[1] - ara[0])/2;
        if (dra[i]<0) { /* Over 0h */
	    dra[i] = adeg(180)-dra[i]; 
	    if (rac[i] < adeg(180)) rac[i] += adeg(180);
	    else                    rac[i] -= adeg(180);
	}
    }

    /* 2 regions defined on RA: select #0 = more restrictive */
    /* Beware limits int4: adeg(360) close to int4 limit:
     * therefore (ra1+ra2)/2 computed with (ra1/2)+(r&2/2) */
    if (dra[0]>dra[1]) {	/* More restrictive region first */
	t = rac[0]; rac[0] = rac[1]; rac[1] = t;
	t = dra[0]; dra[0] = dra[1]; dra[1] = t;
    }

    /* Recompute the limits of the 2 regions in frame centered on rac[0],
     * with intervals in range [-180, +180[  */
    rac[1] -= rac[0]; mrange(ra[3]);
    ra[0] = -dra[0];
    ra[1] =  dra[0];
    ra[2] = rac[1] - dra[1]; if (ra[2] < -adeg(180)) ra[2] += adeg(360);
    ra[3] = rac[1] + dra[1]; if (ra[3] >= adeg(180)) ra[3] -= adeg(360);

    /* Start with closest */
    tested = 0;
    if (rac[1]<0) {
	if (ra[3] < ra[0]) return(0);
	qra[0] = ra[0] + rac[0]; 	/* Lower limit */
	if (ra[3] > ra[1]) {
	    qra[1] = ra[1] + rac[0]; 
	    return(ppmxl_search(qra, gde, photomask, digest));
	}
	/* 2 sub-domains */
	qra[1] = ra[3] + rac[0];
	tested = ppmxl_search(qra, gde, photomask, digest);
	if ((ra[2]>0) && (ra[2]<ra[1])) {
	    qra[0] = ra[2] + rac[0];
	    qra[1] = ra[1] + rac[0];
	    tested += ppmxl_search(qra, gde, photomask, digest);
	}
    }
    else {
	if (ra[2] > ra[1]) return(0);
	qra[1] = ra[1] + rac[0];	/* Upper limit */
	if (ra[2] < ra[0]) {
	    qra[0] = ra[0] + rac[0];
	    return(ppmxl_search(qra, gde, photomask, digest));
	}
	/* 2 sub-domains */
	qra[0] =  ra[2] + rac[0];
	tested = ppmxl_search(qra, gde, photomask, digest);
	if ((ra[3]<0) && (ra[3]>ra[0])) {
	    qra[0] = ra[0] + rac[0];
	    qra[1] = ra[3] + rac[0];
	    tested += ppmxl_search(qra, gde, photomask, digest);
	}
    }
    return(tested) ;
}
#else
#define ppmxl_pos(a,d)	ppmxl_search(a, d, photomask, digest)
#endif

int4 ppmxl_center(double center[2])
/*++++++++++++++++
.PURPOSE  Search PPMXL stars from a central position:
	- either within circle of radius (degrees)
	- or within box of half-dimensions boxy
.RETURNS  Number of tested records.
-----------------*/
{
  int4 ra[2], de[2];
  double value, sr, cosdec, da, dd, maxrad=0;

    ra[0] = 0, ra[1] = adeg(360) - 1 ;
    dd = 180. ;

    if ((radius==0) && (boxy[1]==0)) radius = RADIUS ;
    if (radius > 0) dd = maxrad = radius ;
    if (boxy[1] > 0) 		/* Search in a Box */
	dd = maxrad = sqrt(boxy[0]*boxy[0] + boxy[1]*boxy[1]) ;

    /* Don't accept too large searches... */
    if (dd > 45.) {
	fprintf(stderr, "****Radius [%.5f]deg or Box [%.5fx%.5f] too large\n",
	    maxrad, boxy[0], boxy[1]) ;
	exit(1) ;
    }

    /* Derive first the limits in Dec and RA */
    value = floor((center[1] - dd)*3.6e6); de[0] = value;
    value =  ceil((center[1] + dd)*3.6e6); de[1] = value;
    if (de[0] < de[1]) { value = de[0]; de[0] = de[1]; de[1] = value; }
    if (de[0] >  adeg(90)) de[0] =  adeg(90);
    if (de[1] < -adeg(90)) de[1] = -adeg(90) ;

    cosdec = cosd(center[1]) ;
    sr = sind(maxrad) ;
    du2max = 2. * sind(maxrad/2.) ;

    /* Compute limits on RA */
    if (sr < cosdec)  {		/* Pole not included in circle */
    	da = asind(sr/cosdec) ;
	ra[0] = (center[0] - da)*3.6e6 ;
	if (ra[0] < 0) ra[0] += adeg(360) ;
	ra[1] = (center[0] + da)*3.6e6 ;
	ra[1] %= adeg(360) ;
    }

    /* Define the constants */
    tr_oR(center, localR) ;
    du2max *= du2max ;		/* max. (2*sin(r/2))**2 */

    /* Display the Constraints */
    if (ppmxl_options&1) {
	fprintf(ofile,
	    "#...ppmxl_center(%.5f, %.5f) maxrad=%.5f, limits_mas:\n",
	    center[0], center[1], maxrad) ;
	fprintf(ofile, "#...       ra=[%d,%d]", ra[0], ra[1]) ;
	fprintf(ofile, " de=[%d,%d]\n", de[0], de[1]) ;
    }

    /* Execute the search (no need to merge with RA and Dec limits */
    return(ppmxl_search(ra, de, photomask, digest));
}

/*==================================================================
		Main Program
 *==================================================================*/
#include <bin_struct.h>		/* Decription of the binary structure */

/* Signal: if interrupted, return the signal number */
#include <signal.h>
static char *sig_msg = "";
void OnSignal(signo)
{
  char hostname[32]; time_t now; int secs;
    hostname[0] = 0;
    gethostname(hostname, sizeof(hostname));
    now = time(0); secs = now-started;
    if (signo == SIGALRM) printf(
            "\n#***tooLong %s@%s (>%ds, elapsed=%ds), STOP %s\n",
            theProg, hostname, timeout, secs, sig_msg);
    else printf(
            "\n#***Signal %s@%s#%d (elapsed=%ds), STOP %s\n",
            theProg, hostname, signo, secs, sig_msg);
    exit(signo) ;
}

int main (argc, argv) int argc; char **argv;
{
  char line[BUFSIZ], *p, *a;
  double center[2], flims[2], factor;;
  char *the_center, *next_center, argline[BUFSIZ] ;
  struct s_fields *f ;
  int goon = 1 ;
  int4 tested, aint4, limra[2], limde[2], Jprec[2];
  int i, sign ;
  int non_flagged_arg = 0 ;
  int limradec = 0;
  FILE *input_file ;

    started = time(0);
    theProg = *argv;				/* Prog. name */
    p = thefields[ITEMS(thefields)-1].name;
    p[0] = p[1] = p[2] = p[3] = '\377';		/* LAST FIELD indicator */
    ofile = stdout;

    if (argc < 2) {
	fprintf(stderr, "%s%s", usage, help) ;
	exit(1) ;
    }

    for (i=1; i<15; i++) signal(i, OnSignal) ;

    if (!getenv("PPMXLroot")) {
	if ((p = realdir(*argv))) i = 1;  /* Indicates to remove basename */
	else { p = ROOTdir; i = 0; }
	if (ppmxl_options&1) fprintf(stderr, "#...found PPMXLroot=%s\n", p);
        a = malloc(16+strlen(p));
        sprintf(a, "PPMXLroot=%s", p);
	while ((p = strstr(a, "/./"))) memmove(p, p+2, strlen(p)); 
	if (i) { 
	    i = strlen(a)-1; if (a[i] == '/') a[i--] = 0;
	    while ((i>0) && (a[i] != '/')) i--;
	    if (i>0) a[i] = 0;
	}
        putenv(a);
    }

    /* NOTE: The options can also be  RA-DEC Radius */
    the_center = next_center = (char *)0;
    input_file = stdin ;
    while (--argc > 0) {
	p = *++argv;
	if (ppmxl_options&1) fprintf(stderr, " %s", p);
	if (*p == '-') switch(p[1]) {
	  case 'H':	/* HELP */
	    if (strchr(p, 'b'))
		 printf("\n%s\n#Binary structure of the PPMXL:\n%s",
			 help_bin, bin_struct);
	    else printf("%s", HELP);
	    exit(0) ;
	  case 'h':	/* Help */
	    printf("%s%s", usage, help) ;
	    exit(0) ;
	  case 'f':	/* Next parameter = input file */
	    argc--, argv++;
	    input_file = fopen(*argv, "r") ;
	    if (!input_file) { perror(*argv) ; exit(1) ; }
	    continue ;
	  case 'b':	/* Next parameter = Box limits */
	    argc--, argv++;
	    switch(p[2]) {
	      case 'd': factor = 1; break ;
	      case 's': factor = 1./3600. ; break ;
	      default:  factor = 1./60. ; break ;
	    }
	    i = get_lims(*argv, flims);
	    if (i<0) exit(1);
	    if (i<2) flims[1] = flims[0] ;
	    flims[0] *= factor; flims[1] *= factor;
	    flims[0] /= 2.0 ; flims[1] /= 2.0 ;
	    boxy[0] = flims[0] ; boxy[1] = flims[1] ;
	    thefields[2].lim[1] = thefields[2].factor*tand(flims[0]) ;
	    thefields[3].lim[1] = thefields[3].factor*tand(flims[1]) ;
	    thefields[2].lim[0] = -thefields[2].lim[1] ;
	    thefields[3].lim[0] = -thefields[3].lim[1] ;
	    thefields[2].selected = thefields[3].selected = 1 ;
	    continue ;
	  case 'E':	/* Remove the proper motion effect */
	    optE = 1;
	    continue;
	  case 'R':	/* Root Name of Catalogue  */
	    if (p[2])  p += 2;
	    else p = *++argv, argc-- ;
	    a = malloc(12+strlen(p)) ;
	    sprintf(a, "PPMXLroot=%s", p) ;
	    putenv(a);
	    continue ;
	  case 'r':	/* Next parameter = Radius */
	    switch(p[2]) {
	      case 'd': factor = 1;         break ;
	      case 's': factor = 1./3600. ; break ;
	      case   0:
	      case 'm': factor = 1./60. ;   break ;
	      default:
		fprintf(stderr, "****Bad argument: '%s'\n%s", --p, usage);
		exit(1);
	    }
	    argc--, argv++;
	    i = get_lims(*argv, flims);
	    if (i<0) exit(1);
	    if (i<2) {
		flims[1] = flims[0] ;
		flims[0] = 0 ;
	    }
	    flims[0] *= factor; flims[1] *= factor;
	    radius = flims[1] ;
	    thefields[0].lim[0] = flims[0] < -2147483647.0 ?
	    	-2147483647 : flims[0];
	    thefields[0].lim[1] = flims[1] >  2147483647.0 ?
	    	 2147483647 : flims[1];
	    continue ;
	  case 'm':	/* Max number of records */
	    argc--, argv++;
	    mrec = atoi(*argv) ;
	    if (mrec < 1) mrec = 1999999999;
	    continue ;
	  case 'e':	/* Edit option	*/
	    p += 2;	/* Is the argument of sort glued with -e ? */
	    if (!*p) p = *++argv, argc-- ;
	    if ((p[1]==0) && strchr("mds", *p)); 
	    else opted = 0 ; 
	    if (isdigit(*p)) opted = atoi(p) ;
	    else while(*p) switch(*(p++)) {
		case '.': case ' ': continue ;
		case 'b': opted = (opted&(~(2|4|8)))|2; continue; /* basic */
		case 'd': opted &= ~1 ; continue ;	/* pos. degrees */
		case 's': opted |=  1 ; continue ;	/* pos. sexa    */
		case 'j': case 'J':
		case 'i': opted |=  2 ; continue ;	/* add Jname	*/
		case 'e': opted |=  4 ; continue ;	/* add Epoch	*/
		case 'x': opted |=  8 ; continue ;	/* add x,y	*/
		case 'm': opted |= 16 ; continue ;	/* pos. mas	*/
		case 'p': opted |= 256; continue ;      /* Position only*/
		default:
		   fprintf(stderr, "****Bad argument: -e '%s'\n%s", --p, usage);
		    exit(1) ;
	    }
	    continue ;
	  case 'i':	/* Input ID numbers */
	    input_id = 1 ;
	    if ((argc > 1) && (*argv[1] != '-')) {	/* ID as argument */
		++argv; --argc;
		the_center = strdup(*argv) ;
	    }
	    continue ;
	  case 'l':	/* Set up the limits */
	    argc--, argv++;
	    if (!(f = get_field(thefields, p+2))) {
		fprintf(stderr, "****Unknown field in %s\n%s", p, usage);
		exit(1) ;
	    }
	    f->selected = 1 ;
#if 0	    /* Unused here */
	    if (f->factor == 0) {	/* Type of Survey: S, H, O */
		p = *argv;
		i = toupper(*p); 
		if (!strchr("SHO", i)) fprintf(ofile, 
		    "#***Survey type '%c' not S|H|O", *p);
		f->lim[0] = i;
		p++; if(ispunct(*p)) p++;
		if (!*p) continue;
		i = toupper(*p); 
		if (!strchr("SHO", i)) fprintf(ofile,
			"#***Survey type '%c' not S|H|O", *p);
		f->lim[1] = i;
		continue;
	    }
#endif
	    i = get_lims(*argv, flims);
	    if (i<0) {
		fprintf(stderr, 
			"\n#***Missing or invalid limit spec for: %s\n", p);
		exit(1);
	    }
	    if (i == 3) ;
	    else if (i==1) flims[1] = flims[0] ;
	    else if (i==2) flims[0] = -0x7fffffff ;
	    else if (i==0) flims[0] = -0x7fffffff, flims[1] = 0x7fffffff;
	    /* Convert values to integer.
	       Beware of inan (NaN) values
	    */
	    flims[0] *= f->factor;
	    flims[1] *= f->factor;
	    f->lim[0] = flims[0] < -2147483647.0 ? -2147483647 : flims[0];
	    f->lim[1] = flims[1] >  2147483647.0 ?  2147483647 : flims[1];
	    if (p[2] == 'a') limradec |= 1;
	    else if (p[2] == 'd')
		limradec |= 2;
	    if ((f->lim[0] > f->lim[1]) && (p[2] != 'a'))
		aint4 = f->lim[0], f->lim[0] = f->lim[1], f->lim[1] = aint4 ;
	    continue ;
	  case 's':	/* Sort Order	*/
	    p += 2;	/* Is the argument of sort glued with -s ? */
	    if (!*p) p = *++argv, argc-- ;
	    if (*p == '.') 	/* No Sort at All */
		continue ;
	    for (i = 0 ; *p && (i < ITEMS(compare_fields)); i++, p++) {
		sign = 0 ;
		if (*p == '-') sign = 1, p++ ;
		else if (*p == '+') p++ ;
		if ((a = strchr(p, ','))) *a = 0;
		f = get_field(thefields, p) ;
		if (!f) {
		    fprintf(stderr, "****Unknown field in -s %s\n%s",
		    *argv, usage) ;
		    exit(1) ;
		}
		f = dup_field(f);
	    	f->order    = sign ;
		compare_fields[i++] = f ;
		/* Several sort arguments ? */
		if (!a) break;
		p = a;
		*(p++) = ',';
	    }
	    continue ;
	  case 'c':	/* Get Center	*/
	    argc--, argv++;
	    if (argc < 1) {
		fprintf(stderr, "****Unspecified center in -c argument\n%s",
		    usage) ;
		exit(1) ;
	    }
	    strcpy(argline, *argv);
	    the_center = argline ;
	    while (argc > 1) {		/* Position as several arguments ? */
		if (issign(argv[1][0])) {
		    if (! isdigit(argv[1][1])) break ;
		}
		else if (! isdigit(argv[1][0])) break ;
		argc--, argv++ ;
		strcat(argline, " ") ;
		strcat(argline, *argv) ;
	    }
	    continue ;
	  case 'v':	/* Verbose	*/
	    if (strncmp(p, "-ver", 4) == 0) {
	        fprintf(ofile, "ppmxl(%s) -- Version %s\n", ROOTdir, VERSION);
		exit(0);
	    }
	    ppmxl_options |= 1 ;
	    fprintf(stderr, "#...Options: -v");
	    continue ;
	  case 'o':	/* Specify Output */
	    ++argv, --argc;
	    if (p[2] == 'b') {	/* Binary Output */
		ppmxl_options |= 0x100;
		/* Choose a block which can contain an even number of records */
		mblock = BINSIZE*binlen;
		oblock = (unsigned char *)malloc(mblock);
	    }
	    if (strcmp(*argv, "-") == 0) {
		 /* Comments can't be written on stdout */
		 ofile = stdout;
	    }
	    else ofile = fopen(*argv, "w");
	    if (!ofile) {
		fprintf(stderr, "****PPMXL: ");
		perror(*argv);
		exit(1);
	    }
	    continue;
	  case 'T':	/* TIMEOUT */
	    p += 2;
	    if (isdigit(*p)) timeout = get_delay(p);
	    else { --argc; p = *++argv; timeout = get_delay(p); }
	    continue;
	  case 'w':	/* Whole sky	*/
	    if (strcmp(p, "-whole") == 0) {
		whole_ppmxl = 1 ;
		mrec = 1999999999;
		continue ;
	    }
	    /* NO BREAK */
	  default:
	    fprintf(stderr, "****Unknown argument: %s\n%s", p, usage);
	    exit(1) ;
	}
	/* Non-Option: accept just Jnames  */
	if (toupper(*p) == 'J') {	/* Accept Jnames */
	    input_id = 3;
	    the_center = p;
	    continue;
	}
	if (isdigit(*p) && (non_flagged_arg < 2)) {
	    non_flagged_arg++ ;
	    if (the_center == line) { the_center = *argv; continue ; }
	    i = get_lims(*argv, flims);
	    if (i<0) exit(1);
	    radius = flims[0]/60. ;
	    continue ;
	}
	fprintf(stderr, "****Unknown argument: %s\n%s", p, usage);
	exit(1) ;
    }
    if (ppmxl_options&1) fprintf(stderr, ".\n");

     printf(ppmxl_title, VERSION);

    /* Limits in RA and DEC can be ok */
    if ((input_id == 0) && (limradec == 3) && (!the_center)) {
	input_id = 2;
	the_center = (char *)0;
    }
    else if (!the_center) the_center = line;


    /* Set up the check_fields */
    for (f=thefields, i=0; f->name[0]; f++) {
	if (f->selected == 0) continue;
	check_fields[i++] = f ;
	/* For flags: convert limits into 
	 * [0] = list of bits that MUST be zero
	 * [0] = list of bits that MUST be one  */
	if (f->name[0] == 's') {
	    i = ~(f->lim[0]^f->lim[1]);  /* bits with special meaning  */
	    f->lim[1] = i&f->lim[0];	 /* bits which must be 1 */
	    f->lim[0] = i&(~f->lim[0]);  /* bits which must be 0 */
	}
	/* Flags for fast retrieve of records with specific photometry */
	if (f->name[1] == 0) continue;
	photomask |= f->photomask;
	/*j = strloc(ppmxl_col, f->name[1]);
	if (ppmxl_col[j]) photomask |= 1<<(7-j);
	if (f->name[2] == 0) continue;
	if (f->name[3] == 0) continue;
	j = strloc(ppmxl_col, f->name[3]);
	if (ppmxl_col[j]) photomask |= 1<<(7-j);
	 */
    }

    /* Set up the array of results */
    if (compare_fields[0]) {
#if SORT_METHOD == 2	/* Binary tree */
	arec = (LPPMXLrec *)malloc(mrec*sizeof(LPPMXLrec)) ;
#else
	arec = (PPMXLrec *)malloc(mrec*sizeof(PPMXLrec)) ;
	root = (PPMXLrec **)malloc(mrec*sizeof(PPMXLrec*));
	if (!root) arec = (PPMXLrec *)0;	/* to generate error */
	else *root = 0;
#endif
	if (!arec) {
	    fprintf(stderr,
		"****Requested max.rec. (-m) too large: %d\n",
		mrec);
	    exit(1);
	}
    }

    /* Define the timeout */
    if (!timeout) {
	if ((p=getenv("TIMEOUT"))) timeout = get_delay(p);
	if (!timeout) timeout = TIMEOUT;
    }
    if (timeout>MAX_TIMEOUT) {
	fprintf(stderr, "#+++Timeout specified (%d) limited to %ds.\n",
		MAX_TIMEOUT, timeout);
	timeout = MAX_TIMEOUT;
    }

    while (goon) {
	if (whole_ppmxl) { goon = 0; alarm(100*timeout); }
	else if (the_center != line) 	/* -c was specified ! */
	    goon = 0 ;
	else if (thefields[4].selected && thefields[5].selected)
	    the_center = (char *)0,
	    goon = 0 ;			/* -la and -ld specified */
	else if (the_center == line) {
    	    if (isatty(fileno(input_file)))
		printf("----Give %s: ", prompt[input_id&0xf]);
	    sig_msg = "(reading target)";
	    if (!fgets(line, sizeof(line), input_file)) break ;
	    p = line + strlen(line) ;
	    while ((p>line) && (iscntrl(p[-1]))) p-- ;
	    *p = 0 ;
	    if (strncasecmp(line, "quit", 4) == 0) break ;
	    if (ispunct(line[0])) { Puts(line); continue ; }
	    alarm(timeout); sig_msg = "";
	}
	else goon = 0 ;
    	irec = truncated = tested = matched = 0 ;

	if (whole_ppmxl) {
	    if (optE) printf("%s\n", optEmsg);
	    tested = ppmxl_pos((int4 *)0, (int4 *)0);
	}

	else if (input_id == 0) {	/* From center */
	    ppmxl_options &= ~0x10;
	    if (match_center(the_center, center) < 0) continue ;
	    fprintf(ofile, "#Center: %s\n", the_center) ;
	    if (optE) printf("%s\n", optEmsg);
	    if (ppmxl_options&0x100) ;	/* Binary output: no title */
	    else if (head_issued == 0)  Puts(ppmxl_head(opted)) ;
	    head_issued++;
	    tested = ppmxl_center(center) ;
	}

	else if (input_id == 3) {		/* Get from Jname */
	    ppmxl_options |= 0x10 ;
	    if (match_Jname(the_center, limra, Jprec) < 0) continue;
	    fprintf(ofile, "#Jname: %s\n", the_center);
	    if (optE) printf("%s\n", optEmsg);
	    limde[0] = limra[1];
	    if (limde[0]<0) Jprec[1] = 1-Jprec[1];
	    else            Jprec[1] -= 1;
	    limde[1] = limde[0] + Jprec[1];
	    limra[1] = limra[0] + Jprec[0]-1;
	    tested = ppmxl_pos(limra, limde);
	}

	else if (input_id == 1)  {		/* PPMXL-ID */
	  PPMXLrec rec; int8 id;
	    fprintf(ofile, "#PPMXL %s\n", the_center) ;
	    if (optE) printf("%s\n", optEmsg);
	    id = atoll(the_center);
	    tested = ppmxl_get(id, &rec);
	    if (tested>0) digest(&rec);
	    else if (tested<0) {
#if int8l == 2
		fprintf(ofile, "#***Impossible PPMXL number: %lld\n", id);
#else
		fprintf(ofile, "#***Impossible PPMXL number: %ld\n", id);
#endif
		tested = 0;
	    }
	    else
#if int8l == 2
		fprintf(ofile, "#+++PPMXL-ID does not exist: %lld\n", id);
#else
		fprintf(ofile, "#+++PPMXL-ID does not exist: %ld\n", id);
#endif
	}
	else if (input_id == 2)  {		/* -la and -ld only  */
	    ppmxl_options |= 0x10 ;
	    fprintf(ofile, "#Limits: -la=%d,%dmas -ld=%d,%dmas\n",
		thefields[4].lim[0], thefields[4].lim[1],
		thefields[5].lim[0], thefields[5].lim[1]);
	    if (optE) printf("%s\n", optEmsg);
	    /* Puts(ppmxl_head(opted)) ; Done in ppmxl_pos ! */
	    tested = ppmxl_pos(thefields[4].lim, thefields[5].lim) ;
	}

	/*---- List saved records in order (replace backward by forward link) */
    	if (compare_fields[0]) {
	    if (head_issued == 0) fprintf(ofile, "%s\n", ppmxl_head(opted));
	    head_issued++;
	    print_result(root);
#if SORT_METHOD == 2
	    root = last = (LPPMXrec *)0;
#else
	    if (root) *root = 0;
#endif
	}

	/* In binary output, write the end-of-data */
	if (ppmxl_options&0x100) {
	    ppmxl_flush();
	}

	if (truncated)  {
	    fprintf(ofile, "#+++Truncated to %d out of ", mrec) ;
	    if (matched)  fprintf(ofile, "%d", matched) ;
	    else fprintf(ofile, "(not-computed)") ;
	    fprintf(ofile, " matches (%d tested)\n", tested) ;
	}
	else fprintf(ofile, "#--- %d matches (%d tested)\n",
	    matched, tested);
    }
    ppmxl_close() ;
    exit(0);
}
