/*++++++++++++++
.IDENTIFICATION ppmxl.h
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    CMC-14 Catalogue (Carlsberg Meridian Catalog)
.KEYWORDS       CDS Catalogue Server
.VERSION  1.0   02-Dec-2006
.COMMENTS       Structures & Definitions concerning the CMC-14 Catalog
---------------*/

#ifndef PPMXL_DEF
#define PPMXL_DEF	0

#define DEG	3600000	/* mas in 1 degree */
#include <catutil.h>	/* Includes stdlib, stdio, string, ctype, math */

/*===========================================================================
		Structures
 *===========================================================================*/

/* Note: structure written to be interpretable by the vizSet.awk script */

typedef struct {
    int8 id;  		/* Identifier */
    int4 ra ;		/* (mas) RA(ICRS)  		*/
    int4 de ;		/* (mas) Dec(ICRS)		*/
    int2 e_ra;		/* (mas) Mean errors on ra 	*/ /*=e_RAs */
    int2 e_de;		/* (mas) Mean errors on Dec	*/ /*=e_DEs */
    int4 pmra ;		/* (0.1mas/yr) Proper motion    */
    int4 pmde ;		/* (0.1mas/yr) Proper motion    */
    int2 e_pmra;	/* (0.1mas/yr) Mean error pmra  */
    int2 e_pmde;	/* (0.1mas/yr) Mean error pmde  */ /*---*/
    int2 Nobs;		/* Number of observations       */
    int2 flags;		/* Flags (col#25)               */
    int4 epra;		/* (0.01yr) Epoch in RA         */
    int4 epde;		/* (0.01yr) Epoch in Dec        */
    int2 mag[8];	/* (mmag) J H K b1 b2 r1 r2 i   */
    int2 e_mag[3];	/* (mmag) Error on J H K only   */
    char survey[5];	/* 5 digits for b1 b2 r1 r2 i   */
    char fill;		/* 0x01 = position orig.Epoch computed */
    			/*------------------------------*/
    int4 rho, theta ;	/* (mas) Distance from center 	*/
    int4 xy[2] ;	/* (mas) Values of proj. x,y  	*/
    int4 pmt;           /* Total proper motion 0.1mas/yr*/
} PPMXLrec ;
#define NULLxy	(-999999999)	/* Non-computed values for xy	*/
#define NULL2	(-32768)	/* NULL value for int2 values   */
#define PPMXL_len  256		/* Length edited record...      */
static char ppmxl_col[] = "JHKbbrri";
/* static char ppmxl_set[] = "HOS";     */

/*===========================================================================
		Prototypes Declarations
 *===========================================================================*/

int   ppmxl_init   _ARG((char *PPMXLroot));	/* Default is $PPMXLroot */
/* int   ppmxl_limde  _ARG((int4 dec_lim[2]));	-- Dec. limits of PPMXL14 */
int   ppmxl_close  _ARG((void));
		/* Edition option: 1 in sexa, 2=ID, 4=Epoch */
char *ppmxl_a      _ARG((PPMXLrec *pr, int opt)) ;
char *ppmxl_head   _ARG((int opt)) ;
int4  ppmxl_seek   _ARG((int4 ra, int4 sd)) ; 	/* Position in mas */
int4  ppmxl_search _ARG((int4 ra[2], int4 sd[2], int mask_option,
	    		int (*digest_routine)())) ;
int   ppmxl_get    _ARG((int8 id, PPMXLrec *rec)) ;
int   ppmxl_read   _ARG((PPMXLrec *rec)) ;
int   ppmxl_zopen  _ARG((int zone /*0..720, -1=whole sky*/));
#if SWAP
void  ppmxl_swap   _ARG((PPMXLrec *rec)) ;
#else
#define ppm_swap(a)	/* NOSWAP */
#endif
#endif
