/*++++++++++++++
.IDENTIFICATION ppmxl_def.h
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    USNO-A2.0 Catalogue
.KEYWORDS       CDS Catalogue Server
.VERSION  1.0   02-Dec-2006
.COMMENTS       Structures & Definitions concerning the USNO-A2 Catalog
		only for internal usage.
	The compacted record is made 0f 7 bytes = 56 bits with
	Field_index (3) GSC/ACT(1) Q(1) RA(11) SD(22) MB(9) MR(9)
---------------*/

#ifndef PPMXL_DEF_def
#define PPMXL_DEF_def	0
#ifndef _ARG
#ifdef __STDC__
#define _ARG(A)	A    /* ANSI */
#else
#define _ARG(A)	()  /* non-ANSI */
#define const
#endif
#endif

/* Verify 64-bit machine */
#ifndef int4
#define int4 int
#endif	/* int4 */

/*===========================================================================
		Structures 
 *===========================================================================*/

/* Flags in PPMXL */

#define HEADERlen  80
#define RECLEN1    29		/* Fixed-part length */
#define RECLEN     (RECLEN1+12+10)
#define ASTEP      2500		/* Separation of RA offsets */
#define DSTEP     (15*60000)	/* 15' = 0.25deg */
static char HEADERform[] 
       = "#PPMXL(28+)/s00d.bin N=%d n4=%d Noff=%d*2 oDE=%dmas";

/*===========================================================================
		Prototypes Declarations
 *===========================================================================*/

#define binRA(s)   ((s[12]&0x7f)<<24)|(s[13]<<16)|(s[14]<<8)|s[15]
#define binlen(s)  s[1]&0x80 ? reclen1[s[0]] : reclen2[s[0]]
#define starts_by(p,c)  strncmp(p,c,sizeof(c)-1)==0

/*===========================================================================
		Data organisation
	720 files of 0.25deg (15') each
 *===========================================================================*/
/*
    Flags   = 9b (JHKb2b2r1r2i)+1b. large JHK err.
    ID      = 64b
    Flags   = 4b [last field]
    DE(mas) = 20b (15'=0.25deg)
    RA(mas) = 31b			===> 16 bytes
   eRA(mas) = 8b 
   eDE(mas) = 8b 
   pma(.1m) = 16b (-100000/+100000)  Entre -3 et +3"/yr; audela=Xtra
   pmd(.1m) = 16b (-100000/+100000)  Entre -3 et +3"/yr; audela=Xtra
   epma     = 10b (in 0.mas/yr)
   epmf     = 10b (in 0.mas/yr)
   epRA(cyr) = 14b
   epDE(cyr) = 14b			===> 28 bytes
    Nobs       4b (2-17)
    JHK+Err  = 3*24b or 3*32b (Error+1 stored; value 0 => null)
    Umags    = 16b ((mag+5)*100)*10 + SurveyOrigin

     Total = 28 + 4*3 + 5*2 = 50 bytes max. ==> 36Gbytes


    +00.0(8)  = 8 bits flags photometry
    +01.0(1)  = 1 bit large JHK error
    +01.1(63) = Q3pix
    +09.0(4)  = Code (0-15)
    +09.4(20) = Dec
    +12.0(1)  = FREE
    +12.1(31) = RA(mas)
    +16.0(8)  = eRA(mas)
    +17.0(8)  = eDE(mas)
    +18.0(1)  = FREE
    +18.1(5)  = Nobs
    +18.6(17) = pmRA+64000 [>=128000 ==> Xtra]
    +20.7(17) = pmDE+64000 [>=128000 ==> Xtra]
    +23.0(10) = e_pmRA
    +24.2(14) = EpRA-1900 [0.01yr]
    +26.0(10) = e_pmDE
    +27.2(14) = EpDE-1900 [0.01yr]

    +29.0    ===> Suite : JHK (3bytes if existing) b1b2r1r2i (2bytes if exists)


    File Organisation:
    1 = Header line
    2 = xtra4
    3 = offsets
    4 = full file

*/
#endif
