/*++++++++++++++
.IDENTIFICATION ppmxlsub.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    USNO-A2.0 Catalogue
.KEYWORDS       CDS Catalogue Server
.VERSION  1.0   13-Jun-2010
.COMMENTS       Edition of PPMXL data.
---------------*/

#include <ppmxl_def.h>	/* Structure definitions */
#include <ppmxl.h>	/* Includes stdlib, stdio, string, ctype, math */
#include <catutil.h>	/* Includes stdlib, stdio, string, ctype, math */
#include <q3c.h>	/* Q3C pixel used by PPMXL */

#ifndef ROOTdir
#define ROOTdir "/cats/PPMXL"
#endif

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>      /* O_BINARY */
#include <math.h>
#include <sys/types.h>
#if HAVE_MMAP
#include <sys/mman.h>
#endif

#ifndef O_BINARY	/* Required for MS-DOS	*/
#define O_BINARY 0
#endif
#include <stdio.h>

#ifndef sind
#define  sind(x)        sin(x*(M_PI/180.))
#define  cosd(x)        cos(x*(M_PI/180.))
#define  asind(x)       (180./M_PI)*asin(x)
#endif

/*#define PPMXLlen 	256 */

#ifndef VERIFY
#define VERIFY 0
#else
#define VERIFY 1
#endif

/* Options for PPMXL queries.......	*/
int ppmxl_options ;	/* 0x10 for OrigOrder */
static time_t started;
#if 0
#define edra(b,v,o)	((o)&1 ? edsex_ra(b,v,4) : \
			((o)&16? edmas_ra(b,v)   : eddeg_ra(b,v,6)))
#define edde(b,v,o)	((o)&1 ? edsex_de(b,v,3) : \
			((o)&16? edmas_de(b,v)   : eddeg_de(b,v,6)))
#endif

typedef struct {
    char *filename;             /* Full file name       */
    int file;                   /* file handle          */
    unsigned short mask;        /* Restriction mask (unused)*/
    short int zone;		/* Zone, range 0..719     */
    int4 oDE;                   /* Offset of Declination  */
    int4 bytes;                 /* How many bytes in abin */
    int4 pos;                   /* Current position in abin */
    int4 epos;                  /* Where to stop  */
    int next;                   /* What to do when file exhausted:
                                   0 = stop at EOF, (-1 = stop at EOchunk)
                                  >0 = next zone to read */
    unsigned int4 accessno;     /* To know which one to close */
    char headline[HEADERlen];
    unsigned short nxtra4;	/* Number of extra's          */
    unsigned short Noff;	/* Bytes in offset / offRA    */
    int4 records;		/* How many bytes in the file */
#if HAVE_MMAP
#else
    int chunkno;                /* Loaded chunk */
    int mbin;                   /* Bytes allocated for mbin */
#endif
    unsigned char *abin;        /* Actual data, starts by header  */
    int4  *axtra4;
    int4  *offset, *offRA;
} PPMXLfile;


static unsigned int4 tic = 0;   /* Counter to define oldest ... */
static struct s_PPMXLcat {
    int lroot, mroot;
    char *name;                 /* Root name of PPMXL */
    int i0ls, n1dex;            /* Index in a0ls, #a1dex */
    char *a0ls;                 /* List of files */
    int4 *a1dex;                /* 10*chunks index */
    int ifile, mfile; 
    PPMXLfile *file[3];       /* Opened PPMXL files */
} PPMXLcat ;

/* static int swapping = SWAP;  -- Set to 1 or 2 (swap) */
static unsigned char reclen1[256]; /* Actual record length from 1st byte */
static unsigned char reclen2[256]; /* Record length is case JHK error large  */
static struct q3c_prm hprm;

#define MASK(b)   ((1<<(b))-1)

/*==================================================================
		Internal Utilities
 *==================================================================*/

/* Sort routine */
#if 0
static int i4cmp(int4 *a1, int4 *a2) { return(*a1 - *a2); }

static int s2int(int *a1, int *a2)
/*++++++++++++++++
.PURPOSE  Sort routine (called by qsort), 2 integers (index distance)
.RETURNS  <0 0 >0
-----------------*/
{
  int diff ;
    diff = a1[1] - a2[1] ;
    if (!diff) diff = a1[0] - a2[0] ;
    return(diff) ;
}
#endif

static int4 xget4(unsigned char *a, int b, int len, int4 max, int4 *xtra4)
/*++++++++++++++++
.PURPOSE  Get a value made of 'len' bits, extra values being in a dedicate array
.RETURNS  The value
-----------------*/
{
  int4 value;
    value = get_bits(a, b, len);
    if (value <= max) return(value);
    value = xtra4[value-(max+1)];
    return(value);
}

/*==================================================================
		External Utilities
 *==================================================================*/

void sJname(char *str, int4 ra, int4 de, int modJ)
/*++++++++++++++++
.PURPOSE  Convert a RA and Dec into J-name for PPMXL
.RETURNS  [0,max] = standard; >max => saved in header
.REMARKS  modJ&1 ==> add 1 unit to DE ;
          modJ&2 ==> add 1 unit to RA ;
-----------------*/
{
  int4 val;
    val = de; 
    if (val<0) val = -val;
    val = val/1000;	   /* Value in arcsec */
    if (modJ&1) val++;
                 str[14] = '0' + val%10;
    val /= 10;   str[13] = '0' + val%6;
    val /=  6;   str[12] = '0' + val%10;
    val /= 10;   str[11] = '0' + val%6;
    val /=  6;   str[10] = '0' + val%10;
    val /= 10;   str[ 9] = '0' + val;
                 str[ 8] = de<0 ? '-' : '+';

    /* Write out the RA part of Jname */
    val = ra - (ra/3);     /* RA in 1/10000s */
    val /= 1000;
    if (modJ&2) val++;
                 str[7] = '0' + val%10;
                 str[6] = '.';
    val /= 10;   str[5] = '0' + val%10;
    val /= 10;   str[4] = '0' + val%6;
    val /=  6;   str[3] = '0' + val%10;
    val /= 10;   str[2] = '0' + val%6;
    val /=  6;   str[1] = '0' + val%10;
    val /= 10;   str[0] = '0' + val;
}

/*==================================================================
		Initialisation
 *==================================================================*/

static void init(char *root)
/*++++++++++++++++
.PURPOSE  Initialisation of PPMXL access.
.RETURNS  ---
-----------------*/
{
  int len, i;
  char *v;

    if (PPMXLcat.name) return; 		/* Already opened... */
    started = time(0);

    /* Verify the Swapping ! */
    verify_swap();

    /* Prepare the reclen1 array */
    for (i=0; i<256; i++) {
	len = RECLEN1;
	if (i&0x80) len += 3;	/* Jmag */
	if (i&0x40) len += 3;	/* Hmag */
	if (i&0x20) len += 3;	/* kmag */
	if (i&0x10) len += 2;	/* kmag */
	if (i&0x08) len += 2;	/* kmag */
	if (i&0x04) len += 2;	/* kmag */
	if (i&0x02) len += 2;	/* kmag */
	if (i&0x01) len += 2;	/* kmag */
	reclen1[i] = len;
	/* reclen2: additional byte for JHKerr */
	if (i&0x80) len++;
	if (i&0x40) len++;
	if (i&0x20) len++;
	reclen2[i] = len;
	/* Photometry bits (JHKbbrri) */
    }

    /* Choose the root name */
    if (!root) root = getenv("PPMXLroot") ;
    if ((!root)&&((v = getenv("_"))))  root = realdir(v);
    if (!root) root = ROOTdir;
    PPMXLcat.name = root;

    /* Prepare the length */
    v = PPMXLcat.name; i = strlen(v);
    if (v[--i] != '/') i++;
    PPMXLcat.lroot = i+4;
    PPMXLcat.mroot = i+24;
    PPMXLcat.name = malloc(PPMXLcat.mroot);
    strcpy(PPMXLcat.name, v);
    strcpy(PPMXLcat.name+i, "/dat");
    PPMXLcat.name[PPMXLcat.lroot] = 0;

    /* Final initialisations */
    PPMXLcat.ifile = -1;
    PPMXLcat.mfile = ITEMS(PPMXLcat.file);

    if (ppmxl_options&1) 
	fprintf(stderr, "#...ppmxl_init(%s)\n", root);

    return;
}


/*==================================================================
                Operate on a file
 *==================================================================*/

static int file_close(PPMXLfile *gfile)
/*++++++++++++++++
.PURPOSE  Close (an unassign) an opened file
.RETURNS  0=Ok, -1=error
-----------------*/
{
  int st, i;
    if (!gfile) return(0);
#if HAVE_MMAP
    if (gfile->abin) munmap(gfile->abin, gfile->bytes);
#else
    if (gfile->abin) free(gfile->abin);
#endif
    st = close(gfile->file);
    if (gfile->axtra4) free(gfile->axtra4);
    memset(gfile, 0, sizeof(PPMXLfile));
    for (i=0; i<ITEMS(PPMXLcat.file); i++) {
	if (PPMXLcat.file[i] == gfile) PPMXLcat.file[i] = 0;
    }
    free(gfile);
    return(st);
}

static PPMXLfile *file_new(char *filename, int mask_opt)
/*++++++++++++++++
.PURPOSE  Assign a structure for a file
.RETURNS  Assigned file / NULL if failed
.REMARKS  mask_opt: 0x1000 for Catalog-only, 0x7f for PhotoBands
	On successful output, positioned at 1st record
-----------------*/
{
  PPMXLfile *gfile;
  int file, bytes;
  char *p;

    if (ppmxl_options&1) { int dt;
	dt = time(0) - started;
	fprintf(stderr, "#...file_new(%s) [opt=%04X] +%ds\n", 
		filename, mask_opt, dt) ;
    }
    file = open(filename, O_BINARY);
    if (file < 0) return((PPMXLfile *)0);

    bytes = strlen(filename)+1;
    gfile = (PPMXLfile *)malloc(sizeof(PPMXLfile)+bytes);
    memset(gfile, 0, sizeof(PPMXLfile));
    gfile->filename = (char *)(gfile+1); strcpy(gfile->filename, filename);
    gfile->file = file;
    gfile->mask = mask_opt;

#if HAVE_MMAP   /* Simple file mapping... but local copy of modified vals */
    bytes = lseek(file, 0, SEEK_END);
    lseek(file, 0, SEEK_SET);
    gfile->abin = mmap((void *)0, bytes, PROT_READ, MAP_PRIVATE, file, 0);
    if (gfile->abin  == MAP_FAILED) {
        fprintf(stderr, "#***PPMXLfile mmap(%s): ", filename);
        perror("");
    }
    gfile->bytes = bytes;
    memcpy(gfile->headline, (char *)gfile->abin, HEADERlen); 
    gfile->headline[HEADERlen-1] = 0;
    /* How many chunks ? */
    if (strncmp(gfile->headline, HEADERform, 12) == 0) {
	p = gfile->headline;
        if ((p = strstr(p, "N="))) gfile->records = atoi(p+2);
        if ((p = strstr(p, "n4="))) gfile->nxtra4 = atoi(p+3);
        if ((p = strstr(p, "Noff="))) gfile->Noff = atoi(p+5);
        if ((p = strstr(p, "oDE=")))  gfile->oDE  = atoi(p+4);
	gfile->zone = ((90*DEG)+gfile->oDE)/DSTEP;
	/* Allocate a copy of indexes */
	bytes = (gfile->nxtra4 + 2*gfile->Noff)*4;
	gfile->axtra4 = malloc(bytes);
	gfile->offset = gfile->axtra4 + gfile->nxtra4;
	gfile->offRA  = gfile->offset + gfile->Noff;
	memcpy(gfile->axtra4, gfile->abin+HEADERlen, bytes);
	swapa4(gfile->axtra4, bytes/4);
	gfile->pos  = gfile->offset[0];
	gfile->epos = gfile->offset[gfile->Noff-1];
    }
    else p = (char *)0;
    if (!p) {
        fprintf(stderr, "#***PPMXLfile(%s): bad signature!\n", filename);
        file_close(gfile);
        gfile = (PPMXLfile *)0;
    }
#endif

    return(gfile);
}

#if 0
void ppmxl_com(FILE *file)
/*++++++++++++++++
.PURPOSE  Specify the file for comments
.RETURNS  ---
.REMARKS  stdout by default
-----------------*/
{
    fcom = file;
}

int ppmxl_stop(int flag)
/*++++++++++++++++
.PURPOSE  Specify whether reading should stop at
		EOF (1), endOfChunk (2), never(0)
.RETURNS  Previous status
-----------------*/
{
  int ostat;
    ostat = stopat_eof;
    stopat_eof = flag;
    return(ostat);
}
#endif

static int ppmxl_fopen(char *filename, int mask_opt)
/*++++++++++++++++
.PURPOSE  Open a name-specified file
.RETURNS  >=0 = OK(slot in PPMXLcat.file) / -1 = error
.REMARKS  mask_opt: currently unused...
-----------------*/
{
  PPMXLfile *gfile, *new;
  unsigned int4 oldest;
  int i, o;

    if (ppmxl_options&1) { 
        printf("#...ppmxl_fopen(%s) [opt=%04X]\n", filename, mask_opt);
    }

    if (!PPMXLcat.name) init((char *)0) ;
    new = file_new(filename, mask_opt);
    if (!new) {
        fprintf(stderr, "#***ppmxl_fopen(%s): ", filename);
        perror("");
        return (-1) ;
    }

    /* Find a slot in PPMXLcat */
    oldest = 0xffffffff; o = 0;
    for (i=0; (i<ITEMS(PPMXLcat.file)) && PPMXLcat.file[i] ; i++) {
        gfile = PPMXLcat.file[i];
        if (gfile->accessno < oldest) { o = i; oldest = gfile->accessno; }
    }
    if (i == ITEMS(PPMXLcat.file)) {
        file_close(PPMXLcat.file[o]);
        i = o;
    }
    PPMXLcat.ifile = i;
    PPMXLcat.file[i] = new;
    return(i) ;
}

int ppmxl_zopen(int zone)
/*++++++++++++++++
.PURPOSE  Open a specified zone (range 0..719; -1 = open whole sky)
.RETURNS  >=0 = OK, -1 = error
-----------------*/
{
  int i, h; 
    if (!PPMXLcat.name) init((char *)0) ;
    /* Is the zone already open ? */
    if (zone>=0) {
	for (i=0; i<ITEMS(PPMXLcat.file); i++) {
            if (!PPMXLcat.file[i]) continue;
	    if (PPMXLcat.file[i]->zone == zone) break;
	}
        if (i<ITEMS(PPMXLcat.file)) {
            if (ppmxl_options&1) printf("#...ppmxl_zopen(%03d) is slot#%d\n",
                zone, i) ;
            return(i);
	}
    }
    if (zone < 360) { i = 359-zone; h = 's'; if(zone<0) i = 359; }
    else            { i = zone-360; h = 'n'; }
    sprintf(PPMXLcat.name+PPMXLcat.lroot, "/%c%02d%c.bin", h, i>>2, 'a'+(i&3));
    i = ppmxl_fopen(PPMXLcat.name, 0);
    if ((zone<0) && (i>=0)) PPMXLcat.file[i]->next = 1;
    return(i);
}

static int ppmxl_nextfile(int slot)
/*++++++++++++++++
.PURPOSE  Open next PPMXL file in slot
.RETURNS  slot in PPMXLcat.file if successful; -1 for EOF
-----------------*/
{
  PPMXLfile *gfile;
  int new, next, mask;

    /* Close current file */
    if (slot<0) /* Open whole sky */
	return(ppmxl_zopen(-1));
    gfile = PPMXLcat.file[slot];
    if (!gfile) {
	fprintf(stderr, "#***ppmxl: slot#%d not initialized!\n", slot);
	return(-1);
    }
    next = gfile->next;
    mask = gfile->mask;
    if (next == 0) return(-1);	/* No next file */
    file_close(gfile); PPMXLcat.file[slot] = 0;
    new = ppmxl_zopen(next);
    if (new<0) {
        fprintf(stderr, "#***ppmxl: missing zone %d\n", next);
        perror("");
        return (-1);
    }
    /* Save index for next file to read */
    gfile = PPMXLcat.file[new];
    gfile->next += 1;
    if (gfile->next >= 720) gfile->next = 0;

    /* Try to reuse the old index */
    if (!PPMXLcat.file[slot]) { 
	PPMXLcat.file[slot] = gfile; 
	PPMXLcat.file[new] = 0;
	new = slot;
    }

    return(new);
}

#if HAVE_MMAP
#else
static int read_chunk(int chunkno)
/*++++++++++++++++
.PURPOSE  Read a 'chunk' in the file.
.RETURNS  Number of bytes read, -1=Error
-----------------*/
{
  int4 opos;
  int len;

    if (ppmxl_options&1) fprintf(stderr, "#...read_chunk(%03d) in file: %s\n",
	    chunkno, PPMXLcat.name ? PPMXLcat.name : "(nil)");

    if ((!PPMXLcat.name) || (PPMXLcat.line[0] == 0))	/* File not opened */
	return(-1);

    if ((chunkno<0) || (chunkno>=PPMXLcat.Noff)) {
	fprintf(stderr, "****read_chunk(%d): no such chunk in file: %s\n",
	    chunkno, PPMXLcat.name);
	return(-1);
    }

    /* Is this chunk already loaded ? It yes, nothing to do... */
    if (chunkno == PPMXLcat.ira)
	return(0);

    /* Load the whole chunk into memory */
    opos = PPMXLcat.offsetRA[chunkno];
    if (lseek(PPMXLcat.fno, opos, SEEK_SET) < 0) {
	fprintf(stderr, "****");
	perror(PPMXLcat.name);
	return(-1) ;
    }
    len = PPMXLcat.offsetRA[chunkno+1] - PPMXLcat.offsetRA[chunkno];
    if (len > PPMXLcat.mbuf) {	/* Buffer too small ??? */
	PPMXLcat.mbuf = (len|0x7fff)+1;	/* Multiple of 32K */
	if (PPMXLcat.abuf) free(PPMXLcat.abuf);
	PPMXLcat.abuf =  malloc(PPMXLcat.mbuf);
    }
    PPMXLcat.nbuf = read(PPMXLcat.fno, PPMXLcat.abuf, len);
    if (PPMXLcat.nbuf != len) {
        fprintf(stderr, "#**** got %d/%d bytes from ", PPMXLcat.nbuf, len);
        perror(PPMXLcat.name);
        return(-1);
    }
    PPMXLcat.ira = chunkno;
    PPMXLcat.ofile = opos;

    /* Set the RA definitions */
    PPMXLcat.ra0 = chunkno<<23;
    PPMXLcat.ra1 = (chunkno+1)<<23;
    PPMXLcat.ibuf = 0 ;

    return(len);
}
#endif

#if 0
static int4 aseek(int4 RAmas)
/*++++++++++++++++
.PURPOSE  Position the file to specified RA value (in mas)
.RETURNS  The actual byte position in file / -1 for error
-----------------*/
{
  int4 da;
  int n;

    if (!PPMXLcat.line[0])		/* File not opened */
	return(-1);
    if (read_chunk(RAmas>>23) < 0)
	return(-1);

    da = RAmas&MASK(23);

    /* Find position in current buffer */
    if (da>0) {
        n = binlocf(PPMXLcat.abuf, PPMXLcat.nbuf, PPMXLcat.reclen, ra23, da);
        if (n < 0) n = 0 ;
        if (ra23(PPMXLcat.abuf+n) < da) n += PPMXLcat.reclen;
    }
    else n = 0;
    if (n >= PPMXLcat.nbuf) 	/* Nothing found in this chunk */
	return(-1);

    PPMXLcat.ibuf = n;

    return(PPMXLcat.ofile+n);
}
#endif

#if 0
static int zseek(int4 DEmas)
/*++++++++++++++++
.PURPOSE  Seek to specified zone --
.RETURNS  0 / --1
-----------------*/
{
  int stat;

    if (!PPMXLcat.name) init((char *)0) ;
    if (DEmas<0) { name[0] = 's'; DEmas = -DEmas; }
    else           name[0] = 'n';
    sprintf(name+1, "%04d.bin", 750*(DEmas/(DSTEP)));
    if (name[3] == '5') name[3] = '3';

    /* Open file */
    stat = ppmxl_fopen(name, 0);
    return(stat);
}
#endif

/*==================================================================
		Convert the Input Record(s)
 *==================================================================*/

static int ed_rec(PPMXLfile *gfile, PPMXLrec *pr)
/*++++++++++++++++
.PURPOSE  Convert the current compressed PPMXL record 
	  into its standard structure
.RETURNS  -1=Error / 0=OK / 1 = Error (too high)
.REMARKS  Undefined values set to NULL
-----------------*/
{
  unsigned char *bin;
  unsigned int4 val4;
  int i, m, m0, val;

    /* Convert the compacted record */
    bin = gfile->abin + gfile->pos;
    m0  = bin[0]; if (bin[1]&0x80) m0 |= 0x100;
    pr->id = ((bin[1]&0x7f)<<24) | (bin[2]<<16) | (bin[3]<<8) | bin[4]; 
    pr->id <<= 32;
    val4   = (bin[5]<<24) | (bin[6]<<16) | (bin[7]<<8) | bin[8];
    pr->id|= val4;
    pr->ra = binRA(bin);
    pr->de = get_bits(bin+9, 4, 20) + gfile->oDE;
    pr->flags = get_bits(bin+9, 0, 4);
    pr->e_ra  = bin[16];
    pr->e_de  = bin[17];
    pr->Nobs  = get_bits(bin+18, 1, 5); 
    pr->pmra  =    xget4(bin+18, 6, 17, 130000, gfile->axtra4)-65000; 
    pr->pmde  =    xget4(bin+20, 7, 17, 130000, gfile->axtra4)-65000;
    pr->e_pmra= get_bits(bin+23, 0, 10);
    pr->epra  = get_bits(bin+23,10, 14) + 190000;
    pr->e_pmde= get_bits(bin+26, 0, 10);
    pr->epde  = get_bits(bin+26,10, 14) + 190000;

    /* Get magnitudes; error is stored as (val+1) */
    bin += RECLEN1; m = 0x80;
    for (i=0; i<3; i++, m>>=1) {
	if (m0&m) { 	/* Yes, value is present */
	    pr->mag[i] = get_bits(bin, 0, 15) - 5000;
	    pr->e_mag[i] = get_bits(bin, 15, 9) -1;
	    if (m0&0x100) { /* Large JHK error */
		pr->e_mag[i] = get_bits(bin, 15, 17) -1;
		bin++;
	    }
	    bin += 3;
	    if (pr->e_mag[i] < 0) pr->e_mag[i] = NULL2;
	}
	else pr->mag[i] = pr->e_mag[i] = NULL2;
    }
    for (i=0; i<5; i++, m>>=1) {
	if (m0&m) {	/* Yes, value is present */
	    val = get_bits(bin, 0, 16);
	    val =    xget4(bin, 0, 16, 65000, gfile->axtra4);
	    pr->survey[i] = (val%10)|'0';
	    if (pr->survey[i] == '9') pr->survey[i] = '-';
	    pr->mag[i+3]  = ((val/10) - 1000)*10;	/* Convert mmag */
	    bin += 2;
	}
	else pr->survey[i] = '-', pr->mag[i+3] = NULL2;
    }

    /* Not yet computed */
    pr->rho = pr->theta = pr->xy[0] = pr->xy[1] = pr->pmt = NULLxy;

    return(0) ;
}

int ppmxl_read(PPMXLrec *rec)
/*++++++++++++++++
.PURPOSE  Read next record in PPMXL
.RETURNS  0(End) / -1 (Err) / 1(OK)
.REMARKS  When end if a RA file is found, action depends on stopat_eof
	If rec is NULL, no edition.
-----------------*/
{
  PPMXLfile *gfile;
  unsigned char *b;

    if (!PPMXLcat.name) init((char *)0) ;

    /* First Record in File */
    if (PPMXLcat.ifile < 0) {
	/* Cover whole catalog -- save in slot#0 */
	PPMXLcat.ifile = ppmxl_nextfile(-1);
	if (PPMXLcat.ifile<0) return(-1);
    }
    gfile = PPMXLcat.file[PPMXLcat.ifile];
    if (!gfile) {
	fprintf(stderr, "#+++ppmxl_read: over last record?\n");
	return(0);
    }


    /* Buffer exhausted ? */
    if (gfile->pos >= gfile->epos) {
        if (gfile->next == 0)     /* Don't look for other files */
            return(0);
        if (ppmxl_nextfile(PPMXLcat.ifile) < 0) 
	    return(-1);
	gfile = PPMXLcat.file[PPMXLcat.ifile];
    }

    /* Get current record */
    b = gfile->abin + gfile->pos;
    ed_rec(gfile, rec);   	  /* Record ready */
    gfile->pos += binlen(b);   /* Prepare next entry */
    gfile->accessno = ++tic;
    return(1);
}

static int edra(char *buf, int4 ra, int opt /* 0=deg, 1=sexa */)
/*++++++++++++++++
.PURPOSE  Edit a RA (unit=mas) into HH:MM:SS.SS
.RETURNS  Number of bytes (13 or 11)
-----------------*/
{
  int i ;
  unsigned int4 value;
    if (opt&16) {       /* Edit in mas */
        i = 10 ;
        buf[--i] = '0' + (ra%10) ;
        for (value=ra/10; value; value /= 10) buf[--i] = '0' + (value%10) ;
        while (i>0) buf[--i] = ' ' ;
        return(10) ;
    }
    if (opt) {          /* Sexagesimal */
        value = ra ;
        value = value - (value/3) ;     /* 2/3 for number in 0.0001s */
        value = (value+50)/100;
        for (i=10; i>8; i--) buf[i] = '0' + value%10, value /= 10 ;
        buf[8] = '.' ;
        buf[7] = '0' + value%10, value /= 10 ;
        buf[6] = '0' + value%6,  value /= 6 ;
        buf[5] = ':' ;
        buf[4] = '0' + value%10, value /= 10 ;
        buf[3] = '0' + value%6,  value /= 6 ;
        buf[2] = ':' ;
        buf[1] = '0' + value%10, value /= 10 ;
        buf[0] = '0' + value ;
        return(11) ;
    }
    else {              /* Caution with 32-bit limit !! */
        /* udeg = 2.5/9 (5/18) */
        value = (ra/18);
        value = 5*value + (5*(ra%18)+8)/18;
        for (i=9; i>3; i--) buf[i] = '0' + value%10, value /= 10 ;
        buf[3] = '.' ;
        buf[2] = '0' + value%10, value /= 10 ;
        buf[1] = '0' + value%10, value /= 10 ;
        buf[0] = '0' + value ;
        return(10) ;
    }
}

static int edec(char *buf, int4 dec, int opt /* 0=deg, 1=sexa */)
/*++++++++++++++++
.PURPOSE  Edit a Declination (unit=mas) into +DD:MM:SS.S
.RETURNS  Number of bytes (13 or 11)
-----------------*/
{
  int i ;
  int4 value ;
    value = dec<0 ? -dec : dec ;
    if (opt&16) {       /* Edit in mas */
        i = 10 ;
        buf[--i] = '0' + (value%10) ;
        for (value/=10; value; value /= 10) buf[--i] = '0' + (value%10) ;
        buf[--i] = dec<0 ? '-' : '+';
        while (i>0) buf[--i] = ' ' ;
        return(10) ;
    }
    buf[0] = dec<0 ? '-' : '+';
    if (opt) {          /* Sexagesimal */
        value = (value + 50)/100;
        buf[10] = '0' + value%10, value /= 10 ;
        buf[9] = '.' ;
        buf[8] = '0' + value%10, value /= 10 ;
        buf[7] = '0' + value%6,  value /= 6 ;
        buf[6] = ':' ;
        buf[5] = '0' + value%10, value /= 10 ;
        buf[4] = '0' + value%6,  value /= 6 ;
        buf[3] = ':' ;
        buf[2] = '0' + value%10, value /= 10 ;
        buf[1] = '0' + value ;
        return(11) ;
    }
    else {
#if 0
        value = 3*value ;               /* 10^6/360000 = 3 * (1 - 2/27) */
        value -= ((2*value)/27) ; 
#else
        value = (value*5 + 2)/18;
#endif
        for (i=9; i>3; i--) buf[i] = '0' + value%10, value /= 10 ;
        buf[3] = '.' ;
        buf[2] = '0' + value%10, value /= 10 ;
        buf[1] = '0' + value ;
        return(10) ;
    }
}

static int eddec(char *buf, int value, int width, int decimals)
/*++++++++++++++++
.PURPOSE  Edit a Number with nd decimals
.RETURNS  Number of bytes (6)
.REMARKS  On input, buf[0] = '+' for sign
-----------------*/
{
  int val = value; int i = width; int nd = decimals;
    val = value ;
    if (val<0) { buf[0] = '-'; val = -val; }
    while (--nd>=0) { buf[--i] = '0'|(val%10); val /= 10 ; }
    buf[--i] = '.';
    if (i>1) { buf[--i] = '0'|(val%10); val /= 10 ; }
    while (val && (i>1)) { buf[--i] = '0'|(val%10); val /= 10 ; }
    if (i>1) { buf[--i] = buf[0]; buf[0] = ' '; }
    while (--i>=0) buf[i] = ' ';
    if (val && (buf[0] == ' ')) { buf[0] = '0'|val; val /= 10 ; }
    if (val) {
        fprintf(stderr, "#***Can't edit with %d.%d: %d\n", width, nd, value);
        for (i=0; i<width; i++) buf[i] = '*';
    }
    return(width) ;
}

static int edmag(char *buf, int mag)
/*++++++++++++++++
.PURPOSE  Edit a Magnitude (3 decimals)
.RETURNS  Number of bytes (6)
-----------------*/
{
  int value ;
    value = mag ;
    buf[0] = buf[1] = buf[3] = buf[4] = buf[5] = ' ' ;
    buf[2] = '.' ;
    if (value == NULL2 ) {	/* NULL */
	 buf[1] = buf[2] = buf[3] = '-';
	 return(6);
    }
    if (value <= -10000) value += 0x10000;
    if (value < 0) {
	value = -value ;
	buf[0] = '-' ;
    }
    buf[5] = '0' + value%10, value /= 10 ;
    buf[4] = '0' + value%10, value /= 10 ;
    buf[3] = '0' + value%10, value /= 10 ;
    buf[1] = '0' + value%10; value /= 10 ;
    if ((buf[0] == ' ') && value) {
	buf[0] = '0' + value%10;
	value /= 10;
    }
    while (value) {  /* <=-10 or >=100 */ 
	buf[5] = buf[4];
	buf[4] = buf[3];
	buf[3] = buf[2];
	buf[2] = buf[1];
	if (buf[0] == '-')      buf[1] = '0' + value%10;
	else { buf[1] = buf[0]; buf[0] = '0' + value%10; }
	value /= 10;
    }

    return(6) ;
}

static int edist(char *buf, int4 dist_mas)
/*++++++++++++++++
.PURPOSE  Edit a value in units of mas in arcsec (with 2 decimals only)
.RETURNS  Number of bytes written
-----------------*/
{
  static char aval[16] ;
  char s, *p, *b ;  int4 x ;
    if (dist_mas < 0) s = '-', x = -dist_mas ;
    else s = 0, x = dist_mas  ;
    x = (x+5)/10 ;		/* Round */
    p = aval + sizeof(aval) ;
    *--p = 0 ;
    *--p = '0' + x%10 ; x /= 10 ;
    *--p = '0' + x%10 ; x /= 10 ;
    *--p = '.' ;
    do { *--p = '0' + x%10, x /= 10 ; }
    while (x) ;
    if (s) *--p = s ;

    for (b = buf, x = 9 - strlen(p) ; --x >= 0; b++) *b = ' ' ;
    while (*p) *(b++) = *(p++) ;
    *b = 0 ;
    return(b-buf) ;
}

/*==================================================================
		Edit a Record
 *==================================================================*/

char *ppmxl_head(int opt /*0=deg, 1=sexa, 2=ID, 4=Ep+src, 8=x,y*/)
/*++++++++++++++++
.PURPOSE  Get the Header for one edition
.RETURNS  The Header, starting by #
-----------------*/
{
  static char idc2[] = { '2', '1' };
  static char buf[PPMXL_len] ;
  char *p; int i, b ;

    p = buf ; b = 0 ;
    if (opt&2) {	/* Edit the IDname */
    	p += sprintf(p, "#%-18s", "PPMXL_ID") ;
    	p += strlen(p) ;
	*(p++) = '|';
    }
    else b = 1, *(p++) = '#' ;

    /* Edit the Position */
    i = edra(p, 0, opt&17) + edec(p, 0, opt&17) - b ;
    b = (i-14)/2 ;	/* 14 = length of text */
    while(--b >= 0) *(p++) = ' ', i-- ;
    if (opt&16) strcpy(p, "RAmas(ICRS)Dec");
    else        strcpy(p, "RA   (ICRS)Dec");
    i -= 14, p += 14;
    while (--i >= 0) *(p++) = ' ';
    *(p++) = ' ' ;

    /* Position only: */
    if (opt&256) {
	*p = 0;
	return(buf);
    }

    /* Proper motions */
    p += sprintf(p, "%8s %8s|", "pmRA", "pmDE");

    /* Epoch */
    if (opt&4) {
	p += sprintf(p, "%7s %7s ", "epRA ", "epDE ");
    }
    strcpy(p, "eRA eDE epma epmd|"); p += strlen(p);

    /* Magnitudes */
    for (i=0; i<3; i++) p += sprintf(p, "  %cmag:+/-   ", ppmxl_col[i]);
    --p; *(p++) = '|';
    for (; i<7; i++) p += sprintf(p, "%c%cmag ", ppmxl_col[i], idc2[i&1]);
    p += sprintf(p, " imag ");
    p[-1] = '|';

    /* Numbers of observations */
    if (opt&4) {
        p += sprintf(p, "bbrri|");
	p += sprintf(p, "%s|", "No");
    }
    p += sprintf(p, " S");

    /* Distance in arcsec */
    if (opt&8) sprintf(p, " ;%9s %9s", "x(\")", "y(\")") ;
    else sprintf(p, " ;%9s", "r(\")") ;
    p += strlen(p) ;

    *p = 0 ;

    return(buf) ;
}

char *ppmxl_a(PPMXLrec *pr, int opt)
/*++++++++++++++++
.PURPOSE  Edit (in a static buffer) one record.
.RETURNS  The edited record [internal buffer]
.REMARKS  opt: 0=deg, 1=sexa, 2=ID, 4=Epoch, 8=(x,y) 16=mas
-----------------*/
{
  static char buf[PPMXL_len] ;
  char *p, *q; int i;

    p = buf ;

    if (opt&2) {	/* Edit the Jname */
#if int8l == 2
	p += sprintf(p, "%19lld", pr->id);
#else
	p += sprintf(p, "%19ld", pr->id);
#endif
	*(p++) = '|';
    }

    /* Edit the Position */
#if VERIFY
    p += sprintf(p, "%10d %10d", pr->ra, pr->de);
#else
    p += edra(p, pr->ra, opt&17) ;
    p += edec(p, pr->de, opt&17) ;
#endif
    if (opt&256) { /* Position only */
	*p = 0;
	return(buf); 
    }
    *(p++) = ' ';

    /* Proper Motions */
#if VERIFY
    p += sprintf(p, "%7d", pr->pmra); *(p++) = ' ';
    p += sprintf(p, "%7d", pr->pmde); *(p++) = ' ';
#else
    *p = '+'; p += eddec(p, pr->pmra, 8, 1); *(p++) = ' ';	/* mas/yr */
    *p = '+'; p += eddec(p, pr->pmde, 8, 1); *(p++) = '|';	/* mas/yr */
#endif

    /* Epochs */
    if (opt&4) {
#if VERIFY
        p += sprintf(p, "%5d", pr->epra-200000); *(p++) = ' ';
        p += sprintf(p, "%5d", pr->epde-200000); *(p++) = ' ';
#else
	*p = ' '; p += eddec(p, pr->epra, 7, 2); *(p++) = ' ';
	*p = ' '; p += eddec(p, pr->epde, 7, 2); *(p++) = ' ';
#endif
    }

    /* Mean errors on position + proper motions */
    p += sprintf(p, "%3d", pr->e_ra); *(p++) = ' ';
    p += sprintf(p, "%3d", pr->e_de); *(p++) = ' ';
#if VERIFY
    p += sprintf(p, "%3d", pr->e_pmra); *(p++) = ' ';
    p += sprintf(p, "%3d", pr->e_pmde); *(p++) = ' ';
#else
    *p = ' '; p += eddec(p, pr->e_pmra, 4, 1); *(p++) = ' ';
    *p = ' '; p += eddec(p, pr->e_pmde, 4, 1); *(p++) = '|';
#endif

    /* Magnitudes */
#if VERIFY
    p += sprintf(p, "%5d %5d", pr->mag[5], pr->mag[6]); *(p++) = ' ';
    for (i=0; i<5; i++) {
	p += sprintf(p, "%5d %3d", pr->mag[i], pr->e_mag[i]); *(p++) = ' ';
    }
#else
    for (i=0; i<3; i++) {
        p += edmag(p, pr->mag[i]); q = p;
        p += edmag(p, pr->e_mag[i]); *(p++) = ' ';
	*q = ':';
    }
    --p; *(p++) = '|';
    for (i=0; i<5; i++) {
	p += edmag(p, pr->mag[i+3]); 
	*(--p) = ' '; p++;
    }
    p[-1] = '|';
#endif

    if (opt&4) {
	for (i=0; i<5; i++) *(p++) = pr->survey[i];
        p += sprintf(p, "|%2d|", pr->Nobs); 
    }
    p += sprintf(p, "%2d ", pr->flags);

    /* Distance in arcsec */
    if (pr->rho >= 0)  {
	*(p++) = ' ' ; *(p++) = ';' ; *(p++) = ' ' ;
	if (opt&8) {	/* Edit x,y */
	    p += edist(p, pr->xy[0]);
	    *(p++) = ' '  ;
	    p += edist(p, pr->xy[1]);
	}
	else p += edist(p, pr->rho) ;
    }

    *p = 0 ;

    return(buf) ;
}


void ppmxl_swap(PPMXLrec *pr)
/*++++++++++++++++
.PURPOSE  Swap the bytes in the record (binary output)
.RETURNS  0/1/2 (type of swap)
.REMARKS  Useful for big-endian machines
-----------------*/
{
  int2 *a2; int4 *a4; int i;
    for (a4 = &pr->ra,   i=6;  --i>=0; a4++) *a4 = swap4(*a4);
    for (a2 = &pr->e_ra, i=16; --i>=0; a2++) *a2 = swap2(*a2);
    for (a4 = &pr->rho,  i=5;  --i>=0; a4++) *a4 = swap4(*a4);
}

/*==================================================================
		Operate on a file
 *==================================================================*/

int ppmxl_close()
/*++++++++++++++++
.PURPOSE  Close completely PPMXL
.RETURNS  0 = OK / -1 = error
-----------------*/
{
  int i;
    for (i=ITEMS(PPMXLcat.file); --i>=0; ) {
	if (PPMXLcat.file[i]) file_close(PPMXLcat.file[i]);
	PPMXLcat.file[i] = 0;
    }
    PPMXLcat.ifile = -1;
    return(0);
}

int ppmxl_seek(int4 RAmas, int4 DEmas/*, int mask*/)
/*++++++++++++++++
.PURPOSE  Seek to specified position
.RETURNS  0 = OK (positioned, next record is RA>=RAmas)
-----------------*/
{
  int zone, ifile, i;
  unsigned char *b;
  PPMXLfile *gfile;

    if (!PPMXLcat.name) init((char *)0) ;
    if (RAmas<0) RAmas += 360*DEG;
    zone = ((90*DEG)+DEmas)/DSTEP;
    ifile = ppmxl_zopen(zone);
    if (ifile<0) return(-1);
    gfile = PPMXLcat.file[ifile];

    /* Seek now until the specified RA found */
    i = binloc(gfile->offRA, gfile->Noff, RAmas);
    if ((i<0) || (i>= gfile->Noff)) return(-1);
    gfile->pos = gfile->offset[i];

    /* Move until the RA is there */
    b = gfile->abin + gfile->pos;
    for (i=0; (binRA(b))<RAmas; i++) b += binlen(b);
    gfile->pos = b - gfile->abin;
    if (gfile->pos > gfile->epos) {
	fprintf(stderr, 
	    "#***PROBLEM in PPMXL bad offsets in file: %s\n",
	    gfile->filename);
	exit(3);
    }
    gfile->accessno = ++tic;

    /* The opened file must be at #0 */
    if (ifile>0) { PPMXLfile *f0;
	f0 = PPMXLcat.file[0];
	PPMXLcat.ifile = 0;
	PPMXLcat.file[0] =  gfile;
	PPMXLcat.file[ifile] = f0;
    }

    return(0) ;
}

/* Routine to check for the correct ID */
static PPMXLrec *check_rec;
static int check_id(PPMXLrec *rec) { 
    /*fprintf(stderr, "#...check_id: %ld/%ld\n", rec->id, check_rec->id);*/
    if (rec->id != check_rec->id) return(0);
    memcpy(check_rec, rec, sizeof(PPMXLrec));
    return(-1);			/* Asks to stop */
}

int ppmxl_get(int8 id, PPMXLrec *rec)
/*++++++++++++++++
.PURPOSE  Seek to position specified by PPM-ID
.RETURNS  1=found, 0=notfound -1=error
.REMARKS  Distance between Q3C index and actual position up to 26" 
	  (rounded here to 30")
-----------------*/
{
  double ra,de,sec;
  int4 ralim[2], delim[2];
  int st;
    /* Range of ID = 0 .. 0x5f ff ff ff ff ff ff ff */
    if (id<0) return(-1);
    if (id >= 0x6000000000000000LL) return(-1);
    /* Initialize the Q3C structure */
    if (hprm.nside == 0) init_q3c1(&hprm,1<<30);
    q3c_ipix2ang(&hprm, id, &ra, &de); sec = 1./cosd(de);
    /* Prepare limits RA & Dec, and checking routine */
    memset(rec, 0, sizeof(rec)); rec->id = id; check_rec = rec;
    de = floor(de*3.6e6); delim[0] = delim[1] = de; 
    delim[0] -= 30000; delim[1] += 30000;     /* Search 30"" around */
    ra = floor(ra*3.6e6); ralim[0] = ralim[1] = ra;
    if (sec>200) 
	st = ppmxl_search(0, delim, 0, check_id);
    else { int dra;
        dra = ceil(30000*sec); 
        ralim[0] -= dra; ralim[1] += dra;
        st = ppmxl_search(ralim, delim, 0, check_id);
    }
    if (st<0) return(st);
    return(rec->id == id);	/* Returns 1 if matched ID */
}

static int prec(PPMXLrec *rec) { printf("%s\n", ppmxl_a(rec, 3)); return(1); }

int4 ppmxl_search(int4 ra[2], int4 de[2], int mask, int (*digest_routine)())
/*++++++++++++++++
.PURPOSE  Search PPMXL stars within range of positions.
.RETURNS  Number of tested records.
.REMARKS  mask: restrict to records with specified photometry
.REMARKS  The supplied digest_routine(PPMXLrec *record) does whatever,
	return -1 if stop asked. Default is just to print out.
	The search is done from middle to borders, according to
	a crude distance estimate between the center and the point
-----------------*/
{
  unsigned char *b, mask1[256];
  PPMXLrec rec ;
  PPMXLfile *gfile;
  int4 *list, *alist=0, *elist, *mlist;
  int4 val, spd[3], dstep;
  int4 tested_records=0;
  int la[2], ld[2], over_0h, i, toN, nlist=0;
  int goon=1, seq_order=0;

    if (!PPMXLcat.name) init((char *)0) ;
    if (!digest_routine) digest_routine = prec ;

    /* Modify the mask (OR between bands) */
    if (ppmxl_options&1) {
	fprintf(stderr, "#...ppmxl_search:");
	if (ra) fprintf(stderr, " ra=%d/%d", ra[0], ra[1]);
	if (de) fprintf(stderr, " de=%d/%d", de[0], de[1]);
	fprintf(stderr, ", mask=%03X\n", mask);
    }
    if (mask) {
	for (i=0; i<256; i++) {
	    mask1[i] = i;
	    if ((i&0x18) && (mask&0x100)) mask1[i] |= 0x18;  /* Any blue band */
	    if ((i&0x06) && (mask&0x200)) mask1[i] |= 0x06;  /* Any red  band */
	    mask1[i] &= mask;
	}
	mask = mask1[mask&0xff];
    }
    else memset(mask1, 0, sizeof(mask1));

    if (!de) {
	seq_order = 1;
	ld[0] = -90*DEG;
	ld[1] =  90*DEG;
    }
    else if ((de[0]>de[1])) {  /* Be sure ranges in Declination ordered */
	ld[0] = de[1];
	ld[1] = de[0];
    }
    else {
	ld[0] = de[0];
	ld[1] = de[1];
    }
    de = ld;
#if 0
    if ((!ra) && (!de)) {  /* Read all...*/
	gfile = file_next(0);
	while (gfile) {
	    for (; gfile->pos < gfile->epos; gfile->pos += binlen(b)) {
	        tested_records++;
	        b = gfile->abin + gfile->pos;
	        if (mask && (mask1[b[0]]!=mask))
		    continue;       /* Ignore non-Catalog entries */
	        ed_rec(gfile, &rec);
	        gfile->accessno = ++tic;
	        if ((digest_routine)(&rec) < 0) { goon = 0; break ; }
	    }
	    gfile = file_next(gfile);
	}
	return(tested_records) ;
    }
#endif

    if (!ra) {		/* no RA limit */
	seq_order = 1;
        la[0] = 0 ;
	la[1] = 360*DEG-1;
    }
    else { la[0] = ra[0]; la[1] = ra[1]; }
    ra = la ;

    if (ra[0]<0) ra[0] += adeg(360);
    if (ra[1]<0) ra[1] += adeg(360);
    if (ra[1]>adeg(360)) ra[1] -= adeg(360);
    over_0h = ra[0] > ra[1];
    spd[0] = adeg(90) + de[0];	/* lower limit */
    spd[1] = adeg(90) + de[1];	/* upper limit */
    if (spd[0]<0) spd[0] = 0;
    if (spd[1]>adeg(180)) spd[1] = adeg(180);
    spd[2] = (spd[0]+spd[1])/2;
    toN = (spd[2]%DSTEP) >= (DSTEP/2);
    nlist = 1+((spd[1]/DSTEP)-(spd[0]/DSTEP));
    if (over_0h) nlist *= 2;
    dstep = toN|seq_order ? DSTEP : -DSTEP;
    tested_records = 0 ;

    /* Allocate the list of zones to read.
     * One zone made of 3 numbers: zone#, ra-min, ra-max */
    alist = malloc(3*4*nlist);
    elist = alist + 3*nlist;
    alist[0] = seq_order ? spd[0] : spd[2];
    alist[1] = ra[0];
    alist[2] = ra[1];
    if (over_0h) { 
	alist[3] = alist[0]; alist[5] = ra[1];
	alist[2] = adeg(360); alist[4] = 0;
	list = alist+6;
    }
    else list = alist+3;
    for (i=1; list<elist; i++) {
	list[0] = alist[0] + dstep*i;
	if ((list[0]<0) || (list[0]>=adeg(180))) break;
	list[1] = ra[0];
	list[2] = ra[1];
	if (over_0h) {
	    list[3] = list[0]; list[5] = ra[1];
	    list[2] = adeg(360); list[4] = 0;
	    list += 6;
	}
	else list += 3;
	if (seq_order) continue;
	if (list>=elist) continue;
	/* Next zone to scan if moving away from center */
	list[0] = alist[0] - dstep*i;
	if ((list[0]<0) || (list[0]>=adeg(180))) break;
	list[1] = ra[0];
	list[2] = ra[1];
	if (over_0h) {
	    list[3] = list[0]; list[5] = ra[1];
	    list[2] = adeg(360); list[4] = 0;
	    list += 6;
	}
	else list += 3;
    }
    elist = list;	/* Gnerally not necessry */
    /* mlist specifies whether check agains Dec is required */
    mlist = elist-6; if (over_0h) mlist -= 6;
    if (mlist<alist) mlist = alist;

    for (list=alist; (list<elist) && goon; list += 3) {
	if (ppmxl_seek(list[1], list[0]-adeg(90)))  /* Failed */
	    continue;
	/* Speed-up: read raw data only (with NULL PPMXLrec pointer)
	   and check RA/Dec limits before full edition
	*/
	gfile = PPMXLcat.file[PPMXLcat.ifile];
	for (; gfile->pos < gfile->epos; gfile->pos += binlen(b)) {
	    tested_records++;
	    b = gfile->abin + gfile->pos;
	    val = binRA(b);
	    if (val > list[2]) break;
	    if (mask1[b[0]]!=mask)
		continue;       /* Ignore non-Catalog entries */
	    if (list>=mlist) {
	        val = get_bits(b+9, 4, 20) + gfile->oDE;
	        if (val < de[0]) continue;
	        if (val > de[1]) continue;
	    }
	    ed_rec(gfile, &rec);
	    gfile->accessno = ++tic;
	    if ((digest_routine)(&rec) < 0) { goon = 0; break ; }
	}
    }

    free(alist);
    return(tested_records) ;
}

/*==================================================================
		Main Program
 *==================================================================*/
#ifdef TEST
#if VERIFY
#define EDITop 16|6	/* mas */
#else
#define EDITop  0|6	/* deg */
#endif
static char usage[] = "\
Usage: ppmxlsub [-eEdit_num_oopt] [binfile|-q]\n\
";

static char help[] = "\
		      -q: search by position\n\
		     -id: search by PPMXL-ID (up to 19-digit)\n\
   compressed_input_zone: one of the files created by ppmxlmake\n\
";
int main (argc, argv) int argc; char **argv;
{
  int opt = EDITop; 	/* Editing option, changed by -e */
  char buf[BUFSIZ], *p ;
  int4 ra, de, qlim[4];
  double q[2];
  int8 id;
  int getpos = 0, stat;
  int infile = -1;
  PPMXLrec rec ;

    ppmxl_options = 1 ;
    if (!getenv("PPMXLroot")) putenv("PPMXLroot=../dat");
    ++argv; argc--;
    if (strncmp(*argv, "-e", 2) == 0) {	/* Editing option */
	p = *(argv++); argc--;
	if (p[2]) opt = atoi(p+2);
	else    { opt = atoi(*(argv++)); argc--; }
    }
    if (argc != 1) {
	fprintf(stderr, "%s%s", usage, help) ;
	exit(1) ;
    }
    if (strcmp(*argv, "-q") == 0) getpos = 2 ;
    else if (strcmp(*argv, "-id") == 0) getpos = 1 ;
    else {
	printf("#===Read PPMXL file: %s\n", *argv);
	infile = ppmxl_fopen(*argv, 0);
	printf("#...             file= %d\n", infile);
	printf("%s\n", ppmxl_head(opt));
	while (ppmxl_read(&rec) >0) {
	    printf("%s\n", ppmxl_a(&rec, opt));
	}
	exit(0);
    }

    while (1) {
	if (isatty(0)) {
    	    if (getpos&2)
		 printf("----Give a position (degrees): ") ;
	    else if (getpos)
		 printf("----Give a PPMXL-ID (19-dig.): ") ;
	    else printf("----Give a PPMXL zone [0-719]: ") ;
    	    if (! fgets(buf, sizeof(buf), stdin)) break ;
	    for (p=buf; isspace(*p); p++) ;
	    if (isgraph(*p)) {
		if (getpos&2) {
		    q[0] = atof(p);
		    while (isgraph(*p)) p++ ;
		    while (isspace(*p)) p++ ;
		    q[1] = atof(p) ;
		    ra = q[0]*3.6e6 ; de = 3.6e6*q[1] ;
		    qlim[0] = ra - 3*60000; qlim[1] = ra + 3*60000;
		    if (qlim[0]<0) qlim[0] += 360*3600000;
		    printf("%s\n", ppmxl_head(opt));
		    stat = ppmxl_seek(ra, de);
		    /*stat = ppmxl_search(qlim, qlim+2, 0, 0); */
		    printf("#...ppmxl_seek(%d, %d) => %d\n", ra, de, stat);
		    continue;
		}
		else if (getpos) {
		    id = atoll(p);
		    stat = ppmxl_id(id);
		    printf("#...ppmxl_id(%ld) => %d\n", id, stat);
		}
		else ppmxl_zopen(atoi(p), 0);
	    }
	    else break;	/* Empty line ==> stop */
	}
	printf("%s\n", ppmxl_head(opt));
	while(1) {
	    if (ppmxl_read(&rec)>0) printf("%s\n", ppmxl_a(&rec, opt));
	    else { printf("====EOF====\n"); break; }
	}
    	if (ppmxl_read(&rec)>0) printf("%s\n", ppmxl_a(&rec, opt)) ;
	else { printf("#===EOF===#\n"); break; }
    }
    ppmxl_close() ;
    exit(0);
}
#endif
