/*
	   Copyright (C) 2004-2008 Sergey Koposov
   
    Author: Sergey Koposov, Sternberg Astronomical Institute
    Email: math@sai.msu.ru 
    http://lnfm1.sai.msu.ru/~math

    This file is part of Q3C.

    Q3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Q3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Q3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <q3c.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

inline q3c_coord_t q3c_dist(q3c_coord_t ra1, q3c_coord_t dec1,
			q3c_coord_t ra2, q3c_coord_t dec2)
{
	/* The FASTEST!!! (and precise) way to compute the distance on the sphere
	 * it uses just 3 !!! computation of trigonometric functions
	 */
	
	q3c_coord_t x, y, z;
	x = q3c_sin ((ra1 - ra2) / 2 * Q3C_DEGRA );
	x *= x;
	y = q3c_sin ((dec1 - dec2) / 2 * Q3C_DEGRA);
	y *= y;
	
	/* Seem to be more precise :) */
	z = q3c_cos ((dec1 + dec2)/2 * Q3C_DEGRA);
	z*=z;
	
	return 2 * q3c_asin (q3c_sqrt (x * (z - y) + y)) * Q3C_RADEG;
}


inline q3c_coord_t q3c_sindist(q3c_coord_t ra1, q3c_coord_t dec1,
				q3c_coord_t ra2, q3c_coord_t dec2)
{
	/* The FASTEST!!! (and precise) way to compute the distance on the sphere
	 * it uses just 3 !!! computation of trigonometric functions
	 */
	q3c_coord_t x, y, z;
	x = q3c_sin ((ra1 - ra2) / 2 * Q3C_DEGRA);
	x *= x;
	y = q3c_sin ((dec1 - dec2) / 2 * Q3C_DEGRA);
	y *= y;
	
	/* Seem to be more precise :) */
	z = q3c_cos ((dec1 + dec2)/2 * Q3C_DEGRA);
	z*=z;
	
	return x * (z - y) + y;
}


void q3c_ang2ipix (struct q3c_prm *hprm, q3c_coord_t ra, q3c_coord_t dec,
			q3c_ipix_t *ipix)
		/* ra in degrees, dec in degrees       */
		/* strictly 0<=ra<360 and -90<=dec<=90 */
{
	q3c_coord_t x0 = 0, y0 = 0, ra1, dec1, tmp0;	
	q3c_ipix_t nside = hprm->nside, *xbits = hprm->xbits,
		*ybits = hprm->ybits, xi, yi, i1;
	char face_num;
	
	if (dec >= 90)
	/* Poles */
	{
		face_num = 0;
		x0 = Q3C_HALF;
		y0 = Q3C_HALF;
		goto END1;
	}
	else if (dec <= -90)
	{
		face_num = 5;
		x0 = Q3C_HALF;
		y0 = Q3C_HALF;
		goto END1;
	}
	
	face_num = q3c_fmod ((ra + 45) / 90, 4);
	/* for equatorial pixels we'll have face_num from 1 to 4 */
	ra1 = Q3C_DEGRA * (ra - 90 * (q3c_coord_t)face_num);
	dec1 = Q3C_DEGRA * dec;
	x0 = q3c_tan (ra1);
	y0 = q3c_tan (dec1) / q3c_cos (ra1);
	face_num++;
	
	if (y0 > 1)
	{
		face_num = 0;
		ra1 = Q3C_DEGRA * ra;
		tmp0 = 1 / q3c_tan (dec1);
#ifdef __USE_GNU
		q3c_sincos (ra1, &x0, &y0);
#else
		x0 = q3c_sin (ra1);
		y0 = q3c_cos (ra1);
#endif
		x0 *= tmp0;
		y0 *= (-tmp0);
		/*x0 = q3c_sin(ra1) / q3c_tan(dec1);*/
		/*y0 = -q3c_cos(ra1) / q3c_tan(dec1);*/
		/* I don't know
		 * Probably I should write (sin(ra)/sin(dec))*cos(dec) to
		 * not loose the precision in the region where dec ~=90deg
		 */
	}
	else if (y0 < -1)
	{
		face_num = 5;
		ra1 = Q3C_DEGRA * ra;
		tmp0 = 1 / q3c_tan (dec1);
#ifdef __USE_GNU
		q3c_sincos (ra1, &x0, &y0);
#else
		x0 = q3c_sin (ra1);
		y0 = q3c_cos (ra1);
#endif
		x0 *= (-tmp0);
		y0 *= (-tmp0);
		/*x0 = -q3c_sin(ra1) / q3c_tan(dec1);*/
		/*y0 = -q3c_cos(ra1) / q3c_tan(dec1);*/
	}
	
	x0 = (x0 + 1) / 2; y0 = (y0 + 1) / 2;
	
	END1:
	
	/* Now I produce the final pixel value by converting x and y values
	 * to bitfields and combining them by interleaving, using the
	 * predefined arrays xbits and ybits
	 */
	
	xi = (q3c_ipix_t)(x0 * nside);
	yi = (q3c_ipix_t)(y0 * nside);

	/* This two following statements are written to handle the
	 * case of upper right corner of base square */
	if (xi == nside)
	{
		xi--;
	}
	if (yi == nside)
	{
  		yi--;
	}
	
	i1 = 1 << (Q3C_INTERLEAVED_NBITS);
	
	
#ifdef Q3C_INT4
	*ipix = ((q3c_ipix_t)face_num) * nside * nside +
		xbits[xi % i1] + ybits[yi % i1];
	/*4byte computation*/
#endif /* Q3C_INT4 */
#ifdef Q3C_INT8
	*ipix = ((q3c_ipix_t)face_num) * nside * nside + xbits[xi % i1] +
		ybits[yi % i1] + (xbits[(xi >> Q3C_INTERLEAVED_NBITS) % i1] +
		ybits[(yi >> Q3C_INTERLEAVED_NBITS) % i1]) * i1 * i1;
	/*8byte computation*/
#endif /* Q3C_INT8 */

	/*fprintf(stdout,"YYY %d %.20Lf %.20Lf\n",face_num,ra,dec);
	BIT_PRINT8(*ipix);
	BIT_PRINT8ix(xi);
	BIT_PRINT8iy(yi);*/
}

/* Initialization of the Q3CUBE structure */
void init_q3c1(struct q3c_prm *hprm, q3c_ipix_t nside)
	/* hprm -- Pointer to main Q3C structure
	 * nside -- Nside parameter (number of quadtree subdivisions)
	 */
{
	int i, k, m, l;
	const q3c_ipix_t nbits = Q3C_INTERLEAVED_NBITS;
	/* Number of bits used when interleaving bits
	 * so the size of each allocated array will be 2^16
	 */
	q3c_ipix_t *xbits, *ybits, *xbits1, *ybits1, xybits_size = 1 << nbits;
	hprm->nside = nside;
	xbits = malloc((xybits_size) * sizeof(q3c_ipix_t));
	hprm->xbits = xbits;
	ybits = malloc((xybits_size) * sizeof(q3c_ipix_t));
	hprm->ybits = ybits;
	xbits1 = malloc((xybits_size) * sizeof(q3c_ipix_t));
	hprm->xbits1 = xbits1;
	ybits1 = malloc((xybits_size) * sizeof(q3c_ipix_t));
	hprm->ybits1 = ybits1;

	xbits[0] = 0;
	xbits[1] = 1;
	ybits[0] = 0;
	ybits[1] = 2;
	/*
	BIT_PRINT8(xbits[0]);
	BIT_PRINT8(xbits[1]);
	*/
	for(i = 2, m = 1; i < xybits_size; i++)
	{
		k = i / m;
		if (k == 2)
		{
			xbits[i] = xbits[i / 2] * 4;
			ybits[i] = 2 * xbits[i];
			m *= 2;
			/*
			BIT_PRINT8(xbits[i]);
			fprintf(stdout,"%lld\n",ybits[i]);
			*/
			continue;
		}
		else
		{
			xbits[i] = xbits[m] + xbits[i % m];
			ybits[i] = 2 * xbits[i];
			/*
			BIT_PRINT8(xbits[i]);
			fprintf(stdout,"%lld\n",ybits[i]);
			*/
			continue;
		}
	}
	xbits1[0] = 0;
	xbits1[1] = 1;
	
	/*
	fprintf(stdout,"%lld\n",xbits1[0]);
	fprintf(stdout,"%lld\n",xbits1[1]);
	BIT_PRINT8(xbits[0]);
	BIT_PRINT8(xbits[1]);
	*/
	
	for(i = 2, m = 2, l = 2; i < xybits_size; i++)
	{
		k = i / m;
		
		if (k < 2)
		{
			xbits1[i] = xbits1[i - m];
		}
		else
		{
			if (k == 4)
			{
				xbits1[i] = xbits1[0];
				m *= 4;
				l *= 2;
			}
			else
				xbits1[i] = xbits1[i - 2 * m] + l;
		}
		/* fprintf(stdout,"%lld\n",xbits1[i]); */
	}
	
	ybits1[0] = 0; ybits1[1] = 0;
	/*
	fprintf(stdout,"%lld\n",ybits1[0]);
	fprintf(stdout,"%lld\n",ybits1[1]);
	*/
	
	for(i = 2, m = 1, l = 1; i < xybits_size; i++)
	{
		k = i / m;
		
		if (k < 2)
		{
			ybits1[i] = ybits1[i - m];
		}
		else
		{
			if (k == 4)
			{
				ybits1[i] = ybits1[0];
				m *= 4;
				l *= 2;
			}
			else
				ybits1[i] = ybits1[i - 2 * m] + l;
		}
	 /*fprintf(stdout,"%lld\n",ybits1[i]);*/
	}
}

void q3c_ipix2ang(struct q3c_prm *hprm, q3c_ipix_t ipix,
				  q3c_coord_t *ra, q3c_coord_t *dec)
{
	q3c_ipix_t nside = hprm->nside, ipix1, *xbits1=hprm->xbits1,
			   *ybits1 = hprm->ybits1, i2, i3, x0, y0;
	
	q3c_coord_t x, y, ra0;
	char face_num = ipix / (nside * nside);
	const q3c_ipix_t i1 = 1 << Q3C_INTERLEAVED_NBITS;
	const q3c_ipix_t ii1 = 1 << (Q3C_INTERLEAVED_NBITS / 2);
	ipix1 = ipix % (nside * nside);
	
#ifdef Q3C_INT4 
	i3 = ipix1 % i1;
	i2 = ipix1 / i1;
	x0 = xbits1[i3];
	y0 = ybits1[i3];
	i3 = i2 % i1;
	i2 = i2 / i1;
	x0 += xbits1[i3] * ii1;
	y0 += ybits1[i3] * ii1;
#endif /* Q3C_INT4 */
	
#ifdef Q3C_INT8
	i3 = ipix1 % i1;
	i2 = ipix1 / i1;
	x0 = xbits1[i3];
	y0 = ybits1[i3];
	i3 = i2 % i1;
	i2 = i2 / i1;
	x0 += xbits1[i3] * ii1;
	y0 += ybits1[i3] * ii1;
	i3 = i2 % i1;
	i2 = i2 / i1;
	x0 += xbits1[i3] * ii1 * ii1;
	y0 += ybits1[i3] * ii1 * ii1;	
	i3 = i2 % i1;
	i2 = i2 / i1;
	x0 += xbits1[i3] * ii1 * ii1 * ii1;
	y0 += ybits1[i3] * ii1 * ii1 * ii1;
	/*
	BIT_PRINT8(ipix);
	BIT_PRINT8ix(x0);
	BIT_PRINT8iy(y0);
	*/
#endif /* Q3C_INT8 */
	
	x = (((q3c_coord_t)x0) / nside) * 2 - 1;
	y = (((q3c_coord_t)y0) / nside) * 2 - 1;
	/* Now -1<x<1 and -1<y<1 */
	
	/*fprintf(stdout,"%Lf %Lf %Lf %Lf\n",(x+1)/2,(y+1)/2,x,y);*/
	
	
	if ((face_num >= 1) && (face_num <= 4))
	{
		ra0 = q3c_atan(x);
		*dec = Q3C_RADEG * q3c_atan(y * q3c_cos(ra0));
		ra0 = ra0 * Q3C_RADEG + ((q3c_coord_t)face_num - 1) * 90;
		if (ra0 < 0)
		{
			ra0 += (q3c_coord_t)360;
		}
		*ra = ra0;
	}
	else
	{
		if (face_num == 0)
		{
			ra0 = Q3C_RADEG * (q3c_atan2(-x, y) + Q3C_PI);
			*dec = Q3C_RADEG * q3c_atan( 1 / q3c_sqrt(x * x + y * y));
			*ra = ra0;
		}
		if (face_num == 5)
		{
			ra0 = Q3C_RADEG * (q3c_atan2(-x, -y) + Q3C_PI);
			*dec = -Q3C_RADEG * q3c_atan(1 / q3c_sqrt(x * x + y * y));
			*ra = ra0;
		}		
	}	
	/*fprintf(stdout,"XXX %d %.20Lf %.20Lf\n",face_num,*ra,*dec);*/
}


