/*
 * Just take ipix<-->pos conversion at CDS
 *
	   Copyright (C) 2004-2008 Sergey Koposov
   
    Author: Sergey Koposov, Sternberg Astronomical Institute
    Email: math@sai.msu.ru 
    http://lnfm1.sai.msu.ru/~math

    This file is part of Q3C.

    Q3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Q3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Q3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <math.h>
#include <stdlib.h>
#ifndef Q3C_COMMON_H
#define Q3C_COMMON_H

/* CDS: use int4 / int8 */
#ifndef int4
#define int4 int
#endif
#ifndef int8
#define int8 long long
#define int8l 2		/* is 1 for 'long', 2 for 'long long' */
#endif

/* If You have not specified the Q3C_INT4 then use the Q3C_INT8 instead */
#ifndef Q3C_INT4 
#ifndef Q3C_INT8
#define Q3C_INT8
#endif /* Q3C_INT8 */
#endif /* Q3C_INT4 */

#ifdef Q3C_INT8		/* Define evrything in 64-bit -- pixel ~1mas */
#ifndef q3c_ipix_t
#define q3c_ipix_t int8
#endif /* q3c_ipix_t */

#ifndef Q3C_IPIX_FMT
# if int8l > 1
# define Q3C_IPIX_FMT "%lld"
# else
# define Q3C_IPIX_FMT "%ld"
# endif /* int8l > 1 */
#endif /* Q3C_IPIX_FMT */
#endif   /* Q3C_INT8 */

#ifdef Q3C_INT4		/* Define evrything in 32-bit -- pixel ~1' */
#ifndef q3c_ipix_t
/*#define q3c_ipix_t long int
  typedef unsigned long int q3c_ipix_t;*/
/*typedef uint32 q3c_ipix_t;*/
#define q3c_ipix_t int4
#endif /* q3c_ipix_t */
#ifndef Q3C_IPIX_FMT
#define Q3C_IPIX_FMT "%d"
#endif /* Q3C_IPIX_FMT */  
#endif /* Q3C_INT4 */

/* If You have not specified the Q3C_LONG_DOUBLE macro then we will use simple 
   double functions */
#ifndef Q3C_LONG_DOUBLE
typedef double q3c_coord_t;
#define q3c_pow(a,b) pow(a, b)
#define q3c_fabs(a) fabs(a)
#define q3c_fmod(a,b) fmod(a, b)
#define q3c_cos(a) cos(a)
#define q3c_sin(a) sin(a)
#define q3c_asin(a) asin(a)
#define q3c_acos(a) acos(a)
#define q3c_sincos(a, b, c) sincos(a, b, c)
#define q3c_tan(a) tan(a)
#define q3c_atan(a) atan(a)
#define q3c_atan2(a,b) atan2(a, b)
#define q3c_sqrt(a) sqrt(a)
#define q3c_log(a) log(a)
#define q3c_ceil(a) ceil(a)
#define Q3C_HALF 0.5
#define Q3C_COORD_FMT "%f"
#define Q3C_PI 3.1415926535897932384626433832795028841968
#define Q3C_2PI 6.2831853071795864769252867665590057683936
#define Q3C_PI_2 1.5707963267948966192313216916397514420984
#define Q3C_DEGRA 0.0174532925199432957692369076848861271344
#define Q3C_RADEG 57.2957795130823208767981548141051703324122
#define Q3C_LG2 0.6931471805599453094172321214581765680755
/* 
I really do not see the reason to set them as const ... 
const q3c_coord_t Q3C_PI=3.1415926535897932384626433832795028841968;
const q3c_coord_t Q3C_2PI=6.2831853071795864769252867665590057683936;
const q3c_coord_t Q3C_PI_2=1.5707963267948966192313216916397514420984;
const q3c_coord_t Q3C_DEGRA=0.0174532925199432957692369076848861271344;
const q3c_coord_t Q3C_RADEG=57.2957795130823208767981548141051703324122;
const q3c_coord_t Q3C_LG2=0.6931471805599453094172321214581765680755;
*/
#else /* Q3C_LONG_DOUBLE */
typedef long double q3c_coord_t;
#define q3c_pow(a,b) powl(a, b)
#define q3c_fabs(a) fabsl(a)
#define q3c_fmod(a,b) fmodl(a, b)
#define q3c_cos(a) cosl(a)
#define q3c_sin(a) sinl(a)
#define q3c_asin(a) asinl(a)
#define q3c_acos(a) acosl(a)
#define q3c_sincos(a, b, c) sincosl(a, b, c)
#define q3c_tan(a) tanl(a)
#define q3c_atan(a) atanl(a)
#define q3c_atan2(a,b) atan2l(a, b)
#define q3c_sqrt(a) sqrtl(a)
#define q3c_log(a) logl(a)
#define q3c_ceil(a) ceill(a)
#define Q3C_HALF 0.5L
#define Q3C_COORD_FMT "%.15Lf"
#define Q3C_PI 3.1415926535897932384626433832795028841968L
#define Q3C_2PI 6.2831853071795864769252867665590057683936L
#define Q3C_PI_2 1.5707963267948966192313216916397514420984L
#define Q3C_DEGRA 0.0174532925199432957692369076848861271344L
#define Q3C_RADEG 57.2957795130823208767981548141051703324122L
#define Q3C_LG2 0.6931471805599453094172321214581765680755L
#endif /* Q3C_LONG_DOUBLE */


#ifndef Q3C_CBITS
#define Q3C_CBITS sizeof(q3c_coord_t)
#endif /* Q3C_CBITS */

#ifndef Q3C_IBITS
#define Q3C_IBITS sizeof(q3c_ipix_t)
#endif /* Q3C_IBITS */

#ifndef Q3C_INTERLEAVED_NBITS 
#define Q3C_INTERLEAVED_NBITS 16
#endif /* Q3C_INTERLEAVED_NBITS */


/* this probably will work only in the case of long double variables ?? */

/*#define Q3C_PI    3.1415926535897932384626433832795029L
#define q3c_2_PI  3.1415926535897932384626433832795029L
#define Q3C_2PI 6.2831853071795864769252867665590058L
#define Q3C_PI_2  1.5707963267948966192313216916397514L
#define Q3C_DEGRA 0.01745329251994329576923690768488612L
#define Q3C_RADEG 57.295779513082320876798154814105170L
#define Q3C_LG2   0.301029995663981195213738894724493026768L
#define Q3C_LG2   0.6931471805599453094172321214581765680755L*/
#define Q3C_BOX_INTERSECT(x0, x1, y0, y1, a0, a1, b0, b1) ((((x0 < a0) && \
                                                          (x1 >= a0)) || \
                                                          (x0 <= a1)) && \
                                                          (((y0 < b0) && \
                                                          (y1 >= b0)) || \
                                                          (y0 <= b1)))
/* True if square(x0,x1,y0,y1) and square(a0,a1,b0,b1) have any common point*/ 

#define Q3C_INTERSECT(a, b, x, y) (((a <= y) && (a >= x)) || \
                                  ((a < x) && (b >= x)))

#define Q3C_DISJUNCT 0	/* disjunct areas */
#define Q3C_PARTIAL 1	/* partial coverage */
#define Q3C_EDGE 1		/* point lie on the edge of the region */
#define Q3C_COVER 2		/* the point or the region is completely inside other */


#define Q3C_MAX_N_POLY_VERTEX 100 /* Maximal number of vertices in the polygon */

struct q3c_prm 
{
    q3c_ipix_t nside;
    q3c_ipix_t *xbits;
    q3c_ipix_t *ybits;
    q3c_ipix_t *xbits1;
    q3c_ipix_t *ybits1;
};

struct q3c_square 
{
    q3c_ipix_t x0, y0; /* Integer coordinates of the center of the square for */
    int nside0;       /* the nside0 segmentation */
    char status;
};

/* Declarations of Functions */
void init_q3c1(struct q3c_prm *, q3c_ipix_t);
void q3c_ang2ipix(struct q3c_prm *, q3c_coord_t, q3c_coord_t, q3c_ipix_t *);
void q3c_ipix2ang(struct q3c_prm *, q3c_ipix_t , q3c_coord_t *, q3c_coord_t *);

#endif/*  Q3C_COMMON_H  */
